
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.context.session_context import SessionContext
import logging
import re
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_scripting.common.tools.variable_manager import VariableManager
from holado_core.common.tools.tools import Tools
from holado_value.common.tables.value_table import ValueTable
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_core.common.tables.table import Table
from holado_core.common.tables.table_row import TableRow
from holado_core.common.tables.table_cell import TableCell
from holado_value.common.tables.value_table_row import ValueTableRow
from holado_value.common.tables.value_table_cell import ValueTableCell
from holado_value.common.tools.value_types import ValueTypes
from holado_value.common.tables.comparators.table_2_value_table_with_header_comparator import Table2ValueTable_WithHeaderComparator
from holado_value.common.tables.comparators.table_2_value_table_comparator import Table2ValueTable_Comparator
from holado_core.common.tables.table_manager import TableManager
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.holado_config import Config
from holado_python.standard_library.typing import Typing
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado_python.common.tools.datetime import DateTime, FORMAT_DATETIME_ISO,\
    FORMAT_TIME_ISO
from holado.common.handlers.undefined import default_value

logger = logging.getLogger(__name__)


#TODO: make it a service in SessionContext
class ContextTools(object):
    
    @classmethod
    def format_context_period(cls, context, format_precision_nsec=None, dt_ref=None, use_compact_format=default_value):
        dt_start, dt_end = context.start_datetime, context.end_datetime
        
        # Prepare format of start datetime
        if format_precision_nsec is not None:
            dt_start = DateTime.truncate_datetime(dt_start, precision_nanoseconds=format_precision_nsec)
        dt_format_start = cls.get_datetime_format_compared_to_reference(dt_start, dt_ref, use_compact_format=use_compact_format)
        
        # Prepare format of end datetime
        if dt_end is not None:
            if format_precision_nsec is not None:
                dt_end = DateTime.truncate_datetime(dt_end, precision_nanoseconds=format_precision_nsec)
            dt_format_end = cls.get_datetime_format_compared_to_reference(dt_end, dt_start, use_compact_format=use_compact_format)
            if len(dt_format_end) > len(dt_format_start):
                dt_format_start = dt_format_end
        else:
            dt_format_end = None
        
        # Format datetimes
        start_txt = DateTime.datetime_2_str(dt_start, dt_format=dt_format_start)
        end_txt = DateTime.datetime_2_str(dt_end, dt_format=dt_format_end) if dt_end is not None else ''
        
        # Truncate formatted datetimes if needed
        if format_precision_nsec is not None:
            trunc_len = len(f'{int(format_precision_nsec)}') - 4
            if trunc_len > 0:
                start_txt = start_txt[:-trunc_len-1]+'Z' if start_txt.endswith('Z') else start_txt[:-trunc_len]
                if len(end_txt) > 0:
                    end_txt = end_txt[:-trunc_len-1]+'Z' if end_txt.endswith('Z') else end_txt[:-trunc_len]
        
        return f"[{start_txt} - {end_txt}]"
    
    @classmethod
    def get_datetime_format_compared_to_reference(cls, dt, dt_ref=None, use_compact_format=default_value):
        if use_compact_format is default_value:
            use_compact_format = Config.report_compact_datetime_period  # @UndefinedVariable
        
        if dt_ref is None:
            return FORMAT_DATETIME_ISO
        
        if not use_compact_format or dt.date() != dt_ref.date():
            return FORMAT_DATETIME_ISO
        elif dt.hour != dt_ref.hour:
            return FORMAT_TIME_ISO
        elif dt.minute != dt_ref.minute:
            return '%M:%S.%f'
        elif dt.second != dt_ref.second:
            return '%S.%f'
        else:
            return '.%f'
    







# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
import base64
import codecs
import logging
from holado_core.common.tables.table import Table
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_value.common.tables.value_table import ValueTable
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing
from holado_value.common.tables.value_table_manager import ValueTableManager

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_unique_value_manager():
    return SessionContext.instance().unique_value_manager

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()





@Given(r"(?P<var_name>{Variable}) = convert list (?P<list_var_name>{Variable}) to list of object attribute (?P<attribute_name>{Str})")
def step_impl(context, var_name, list_var_name, attribute_name):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    attr_name = StepTools.evaluate_scenario_parameter(attribute_name)
    list_obj = StepTools.evaluate_variable_value(list_var_name)
    
    res = []
    for index, obj in enumerate(list_obj):
        if hasattr(obj, attr_name):
            attr_val = getattr(obj, attr_name)
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Result list - add attribute value [{attr_val}] (type: {Typing.get_object_class_fullname(attr_val)} ; dir: {dir(attr_val)})")
            res.append(attr_val)
        else:
            raise FunctionalException(f"In list, object of index {index} hasn't attribute '{attr_name}'")
    
    __get_variable_manager().register_variable(var_name, res)
    
@Step(r"(?P<var_name>{Variable}) = convert object value (?P<object_value>{Str}) to string")
def step_impl(context, var_name, object_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    object_value = StepTools.evaluate_variable_value(object_value)
    if isinstance(object_value, bytes):
        res = object_value.decode('utf-8')
    else:
        res = str(object_value)
    __get_variable_manager().register_variable(var_name, res)
    
@Step(r"(?P<var_name>{Variable}) = convert object value (?P<object_value>{Str}) to integer")
def step_impl(context, var_name, object_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    object_value = StepTools.evaluate_variable_value(object_value)
    __get_variable_manager().register_variable(var_name, int(object_value))
    
@Step(r"(?P<var_name>{Variable}) = convert object value (?P<object_value>{Str}) to list")
def step_impl(context, var_name, object_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    object_value = StepTools.evaluate_variable_value(object_value)
    
    if isinstance(object_value, Table):
        table = object_value
        if table.nb_columns > 1 and table.nb_rows > 1:
            raise FunctionalException(f"Table must have only one column or only one row (table has {table.nb_columns} and {table.nb_rows})")
        
        if table.nb_columns > 1 and table.nb_rows == 1:
            row = table.get_row(0)
        elif table.nb_rows > 1 and table.nb_columns == 1:
            row = table.get_column(0)
        else:
            row = None
            
        if row is not None:
            if ValueTableManager.is_value_table(table):
                res = [c.value for c in row]
            else:
                res = [c.content for c in row]
        else:
            res = []
    else:
        res = list(object_value)
    
    __get_variable_manager().register_variable(var_name, res)
    
@Step(r"(?P<var_name>{Variable}) = convert object value (?P<object_value>{Str}) to dict")
def step_impl(context, var_name, object_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    object_value = StepTools.evaluate_variable_value(object_value)
    __get_variable_manager().register_variable(var_name, dict(object_value))

@Step(r"(?P<var_name>{Variable}) = convert object value (?P<object_value>{Str}) to base 64")
def step_impl(context, var_name, object_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    object_value = StepTools.evaluate_variable_value(object_value)
    if isinstance(object_value, bytes):
        value_bytes = object_value
    elif isinstance(object_value, str):
        value_bytes = object_value.encode('utf-8')
    else:
        raise TechnicalException(f"Unexpected object value type {Typing.get_object_class_fullname(object_value)} (allowed types: string, bytes)")
    res = base64.b64encode(value_bytes)
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = convert hex string (?P<object_value>{Str}) to bytes")
def step_impl(context, var_name, object_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    object_value = StepTools.evaluate_scenario_parameter(object_value)
    if isinstance(object_value, str):
        res = codecs.decode(object_value, 'hex_codec')
    else:
        raise TechnicalException(f"Unexpected value type {Typing.get_object_class_fullname(object_value)}")
    __get_variable_manager().register_variable(var_name, res)
    
@Step(r"(?P<var_name>{Variable}) = convert hex string (?P<object_value>{Str}) to base 64")
def step_impl(context, var_name, object_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    execute_steps(u"""
        {keyword} __{var_name}_BYTES__ = convert hex string {object_value} to bytes
        {keyword} {var_name} = convert object value __{var_name}_BYTES__ to base 64
        """.format(keyword=context.step.keyword, var_name=var_name, object_value=object_value))
    

@Step(r"(?P<var_name>{Variable}) = convert base 64 (?P<value>{Str}) to bytes")
def step_impl(context, var_name, value):
    var_name = StepTools.evaluate_variable_name(var_name)
    value = StepTools.evaluate_scenario_parameter(value)
    if not isinstance(value, str):
        raise TechnicalException(f"Unexpected value type {Typing.get_object_class_fullname(value)} (allowed types: string)")
    res = base64.b64decode(value)
    __get_variable_manager().register_variable(var_name, res)
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import logging
from holado_core.common.block.base import BaseScope
from holado_core.common.block.block_steps import BlockSteps
from holado_test.behave.behave import execute_steps
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class ScopeSteps(BaseScope):
    '''
    Scope block that should only contains steps.
    This scope block is used when a new scope (function, for,...) is defined in a scenario.

    When processing this scope, a 'context' is supposed to be passed as parameter to 'process' method
    Warning: it remains possible to add any block, not only steps. Be carefull that added blocks are compatible with parameter 'context' passed to 'process' method.
    '''
    
    def __init__(self, name, ID=None):
        super().__init__(f"steps scope {name}", ID=ID)
        
    def add_steps(self, steps_str):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"In scope '{self.name}', add steps:\n{steps_str}")
        self.add_block( BlockSteps(steps_str) )
    
class ScopeForSteps(ScopeSteps):
    def __init__(self, name, iter_var_name, iterable, ID=None, variable_manager=None, scope_manager=None):
        super().__init__(name, ID=ID)
        
        self.__iter_var_name = iter_var_name
        self.__iterable = iterable
        
        self.__variable_manager = variable_manager
        self.__scope_manager = scope_manager
    
    def process(self, *args, **kwargs):
        """
        Process the scope.
        Returns last scope step result
        """
        res = None
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Processing scope [{self.name}]: begin")

        self._process_start()
        try:
            for index, value in enumerate(self.__iterable):
                execute_steps(u"""
                    Given FOR_INDEX_{iter_var_name} = {index}
                    """.format(iter_var_name=self.__iter_var_name, index=index))
                self.__variable_manager.register_variable(self.__iter_var_name, value)
                
                if self.__scope_manager:
                    level_already_exists = self.__scope_manager.has_scope_level("steps")
                    origin_level = self.__scope_manager.scope_level("steps")
                    if level_already_exists:
                        self.__scope_manager.increase_scope_level("steps")
                
                try:
                    self._process_blocks(*args, **kwargs)
                finally:
                    if self.__scope_manager:
                        self.__scope_manager.reset_scope_level("steps", origin_level)
        finally:
            self._process_end()
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Processing scope [{self.name}]: end")

        return res
    
class ScopeWhileSteps(ScopeSteps):
    def __init__(self, name, condition_expression, ID=None, expression_evaluator=None, scope_manager=None):
        super().__init__(name, ID=ID)
        
        self.__condition_expression = condition_expression
        
        self.__expression_evaluator = expression_evaluator
        self.__scope_manager = scope_manager
    
    def process(self, *args, **kwargs):
        """
        Process the scope.
        Returns last scope step result
        """
        res = None
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Processing scope [{self.name}]: begin")

        self._process_start()
        try:
            n = 0
            _, cond = self.__expression_evaluator.evaluate_python_expression(self.__condition_expression)
            if not (isinstance(cond, bool) or Converter.is_boolean(cond)):
                raise FunctionalException(f"Condition expression is not a boolean: [{self.__condition_expression}] => [{cond}] (type: {Typing.get_object_class_fullname(cond)})")
            
            while isinstance(cond, bool) and cond or Converter.is_boolean(cond) and Converter.to_boolean(cond):
                n += 1
                execute_steps(u"""
                    Given WHILE_ITERATION = {n}
                    """.format(n=n))
                
                if self.__scope_manager:
                    level_already_exists = self.__scope_manager.has_scope_level("steps")
                    origin_level = self.__scope_manager.scope_level("steps")
                    if level_already_exists:
                        self.__scope_manager.increase_scope_level("steps")
                
                try:
                    self._process_blocks(*args, **kwargs)
                finally:
                    if self.__scope_manager:
                        self.__scope_manager.reset_scope_level("steps", origin_level)
                
                _, cond = self.__expression_evaluator.evaluate_python_expression(self.__condition_expression)
                if not (isinstance(cond, bool) or Converter.is_boolean(cond)):
                    raise FunctionalException(f"Condition expression is not a boolean: [{self.__condition_expression}] => [{cond}] (type: {Typing.get_object_class_fullname(cond)})")
        finally:
            self._process_end()
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Processing scope [{self.name}]: end")

        return res
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from datetime import timedelta, datetime

from holado.common.handlers.undefined import default
from holado_security.token.token_manager import TokenManager, ExpiringTokenManager
from holado_security.tools.keycloak.keycloak_client import KeycloakClient


logger = logging.getLogger(__name__)

class KeycloakTokenManager(ExpiringTokenManager):
    """
    Manage Keycloak tokens.
    Note: In this implementation, 'request_kwargs' argument of method get_token has to be arguments of KeycloakClient.get_token.
          If another signature is needed, it can be done by overriding methods 'get_token', 'request_token' and 'request_refresh'.
    """

    def __init__(self):
        super().__init__(with_refresh=True)

    def get_token_type(self, *, uid, request_kwargs=None, needed_time_s=default):
        return super().get_token(uid=uid, request_kwargs=request_kwargs, needed_time_s=needed_time_s)["token_type"]

    def get_access_token(self, *, uid, request_kwargs=None, needed_time_s=default):
        return super().get_token(uid=uid, request_kwargs=request_kwargs, needed_time_s=needed_time_s)["access_token"]

    def get_refresh_token(self, *, uid):
        return super().get_token(uid=uid)["refresh_token"]

    def get_expires_in(self, *, uid):
        return super().get_token(uid=uid)["expires_in"]

    def ensure_token_is_valid(self, *, uid, request_kwargs=None, needed_time_s=default):
        needed_time_s = needed_time_s if needed_time_s is not default else self.default_needed_time_s

        # Ensure needed_time_s is at most expires_in/2 to optimize the number of refresh
        if self.has_token(uid=uid):
            expires_in = self.get_expires_in(uid=uid)
            if expires_in // 2 < needed_time_s:
                needed_time_s = expires_in // 2
        super().ensure_token_is_valid(uid=uid, request_kwargs=request_kwargs, needed_time_s=needed_time_s)

    def request_token(self, *, keycloak_client, **request_kwargs):
        return keycloak_client.get_token(**request_kwargs)

    def request_refresh(self, *, uid, keycloak_client, **request_kwargs):
        kwargs = {k:request_kwargs[k] for k in request_kwargs if k in ('realm_name', 'client_id', 'client_secret')}
        return keycloak_client.get_token_by_grant_type_refresh_token(refresh_token=self.get_refresh_token(uid=uid), **kwargs)

    def _new_token_info(self, *, token, request_kwargs=None):
        res = super()._new_token_info(token=token, request_kwargs=request_kwargs)
        res.expiration_date = res.creation_date + timedelta(seconds=token["expires_in"])
        res.refresh_expiration_date = res.creation_date + timedelta(seconds=token["refresh_expires_in"])
        return res

    def logout(self, *, uid, keycloak_client, **kwargs):
        """ Logout from Keycloak with token of given UID"""
        # Logout from keycloak
        keycloak_client.logout(refresh_token=self.get_refresh_token(uid=uid), **kwargs)

        # Remove stored token
        self.delete_token(uid=uid)






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from builtins import range
from builtins import super

from holado_core.common.tools.comparators.comparator import Comparator
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_core.common.tables.comparators.string_table_row_comparator import StringTableRowComparator
from holado_core.common.exceptions.exceptions import Exceptions
from holado_core.common.tools.tools import Tools
from holado_core.common.tables.comparators.table_row_comparator import TableRowComparator

logger = logging.getLogger(__name__)


class TableComparator(Comparator):
    def __init__(self, row_comparator=None, cell_comparator=None):
        super().__init__("table")
        
        if row_comparator is not None:
            self.__row_comparator = row_comparator
        elif cell_comparator is not None:
            self.__row_comparator = TableRowComparator(cell_comparator)
        else:
            self.__row_comparator = StringTableRowComparator()

    @property
    def row_comparator(self):
        return self.__row_comparator
    
    def equals(self, table_1, table_2, is_obtained_vs_expected=True, raise_exception=True, do_log=True, return_diffs=False):
        """ Return if tables are equal.
        If return_diffs=True and raise_exception=False, also return the indexes of rows with differences.
        """
        try:
            res = True
            res_diffs = []
            
            # Verify number of rows
            if table_1.nb_rows != table_2.nb_rows:
                if raise_exception:
                    msg = f"Tables have not same number of rows (table {self._get_name_1(is_obtained_vs_expected)}: {table_1.nb_rows} rows ; table {self._get_name_2(is_obtained_vs_expected)}: {table_2.nb_rows} rows)"
                    raise VerifyException(self._build_table_compare_error_message(msg, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected))
                else:
                    return False, None if return_diffs else False

            # Compare rows content
            for i in range(table_1.nb_rows):
                try:
                    res = self.__row_comparator.equals(table_1.get_row(i),
                                                       table_2.get_row(i),
                                                       is_obtained_vs_expected=is_obtained_vs_expected,
                                                       raise_exception=raise_exception,
                                                       do_log=do_log)
                except FunctionalException as exc:
                    msg = f"Difference exists at line of index {i}"
                    raise VerifyException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc
                except Exception as exc:
                    msg = f"Error while comparing line of index {i}"
                    raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc

                if not res:
                    res_diffs.append(i)
                    if raise_exception:
                        raise TechnicalException("This case should have already raised an exception")
                    elif not res_diffs:
                        if do_log and Tools.do_log(logger, logging.DEBUG):
                            msg = f"Difference exists at line of index {i}"
                            logger.debug(self._build_table_compare_error_message(msg, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected))
                        break
                    
            return res, res_diffs if return_diffs else res
        except (FunctionalException, TechnicalException) as exc:
            raise exc
        except Exception as exc:
            msg = "Error while comparing tables"
            raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc

    def contains_rows(self, table_1, table_2, check_row_order = True, is_obtained_vs_expected = True, raise_exception = True):
        try:
            res = True
            
            # Compare rows content
            start_index = 0
            for i in range(table_2.nb_rows):
                try:
                    res, row_index = self.contains_row(table_1, table_2.get_row(i), start_row_index = start_index,
                                                       is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception)
                    if check_row_order:
                        start_index = row_index + 1
                except FunctionalException as exc:
                    msg = f"Table {self._get_name_1(is_obtained_vs_expected)} doesn't contain row {self._get_name_2(is_obtained_vs_expected)} of index {i}"
                    raise VerifyException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc
                except Exception as exc:
                    msg = f"Error while checking if table {self._get_name_1(is_obtained_vs_expected)} contains row {self._get_name_2(is_obtained_vs_expected)} of index {i}"
                    raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc
                if not res:
                    break
                    
            return res
        except (FunctionalException, TechnicalException) as exc:
            raise exc
        except Exception as exc:
            msg = f"Error while checking if table {self._get_name_1(is_obtained_vs_expected)} contains rows of table {self._get_name_2(is_obtained_vs_expected)}"
            raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc

    def doesnt_contain_rows(self, table_1, table_2, is_obtained_vs_expected = True, raise_exception = True):
        try:
            res = True
            
            # Compare rows content
            for i in range(table_2.nb_rows):
                try:
                    row_res, _ = self.contains_row(table_1, table_2.get_row(i),
                                                       is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=False)
                except Exception as exc:
                    msg = f"Error while checking if table {self._get_name_1(is_obtained_vs_expected)} contains row {self._get_name_2(is_obtained_vs_expected)} of index {i}"
                    raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc

                if row_res:
                    if raise_exception:
                        msg = f"Table {self._get_name_1(is_obtained_vs_expected)} contains row {self._get_name_2(is_obtained_vs_expected)} of index {i}"
                        raise VerifyException(self._build_table_compare_error_message(msg, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected))
                    else:
                        res = False
                        break
                
            return res
        except (FunctionalException, TechnicalException) as exc:
            raise exc
        except Exception as exc:
            msg = f"Error while checking if table {self._get_name_1(is_obtained_vs_expected)} contains rows of table {self._get_name_2(is_obtained_vs_expected)}"
            raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=table_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc

    def contains_row(self, table_1, row_2, start_row_index = 0, is_obtained_vs_expected=True, raise_exception=True):
        try:
            res = True
            res_index = -1
            
            # Compare rows content
            for i in range(start_row_index, table_1.nb_rows):
                row_1 = None
                try:
                    row_1 = table_1.get_row(i)
                    res = self.__row_comparator.equals(row_1, row_2,
                                                       is_obtained_vs_expected=is_obtained_vs_expected,
                                                       raise_exception=False,
                                                       do_log=False)
                except Exception as exc:
                    msg = f"Error while comparing table {self._get_name_1(is_obtained_vs_expected)} row of index {i}"
                    raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=row_1, input_2=row_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc

                if res:
                    res_index = i
                    break
                
            if not res and raise_exception:
                msg = f"Table {self._get_name_1(is_obtained_vs_expected)} doesn't contain the row {self._get_name_2(is_obtained_vs_expected)}"
                raise VerifyException(self._build_table_compare_error_message(msg, input_1=table_1, input_2=row_2, is_obtained_vs_expected=is_obtained_vs_expected))

            return res, res_index
        except (FunctionalException, TechnicalException) as exc:
            raise exc
        except Exception as exc:
            msg = f"Error while searching row {self._get_name_2(is_obtained_vs_expected)} in table {self._get_name_1(is_obtained_vs_expected)}"
            raise TechnicalException(self._build_table_compare_error_message(msg, exception=exc, input_1=table_1, input_2=row_2, is_obtained_vs_expected=is_obtained_vs_expected)) from exc

    def _represent_table(self, table, indent):
        try:
            return table.represent(indent)
        except Exception as exc:
            return f"[ERROR while representing table: {Exceptions.exception_message(exc)}]"
    
    def _represent_row(self, row, indent):
        try:
            return row.represent(indent)
        except Exception as exc:
            return f"[ERROR while representing table row: {Exceptions.exception_message(exc)}]"

    def __get_input_type_name(self, input):
        from holado_core.common.tables.table import Table
        from holado_core.common.tables.table_row import TableRow

        if isinstance(input, Table):
            return 'table'
        elif isinstance(input, TableRow):
            return 'table'
        else:
            raise TypeError(f"Unsupported input type: {type(input)}")

    def __represent_input(self, input, indent):
        from holado_core.common.tables.table import Table
        from holado_core.common.tables.table_row import TableRow

        if isinstance(input, Table):
            return self._represent_table(input, indent)
        elif isinstance(input, TableRow):
            return self._represent_row(input, indent)
        else:
            raise TypeError(f"Unsupported input type: {type(input)}")

    def _build_table_compare_error_message(self, msg, error_msg=None, exception=None, input_1=None, input_2=None, is_obtained_vs_expected=True):
        msg_lines = [msg]

        if error_msg is not None:
            msg_lines.append(f" -> {error_msg}")
        elif exception is not None:
            msg_lines.append(f" -> {Tools.indent_string(4, Exceptions.exception_message(exception), do_indent_first_line=False)}")

        if input_1 is not None or input_2 is not None:
            msg_lines.append("---- PARAMETERS:")
            if input_1 is not None:
                msg_lines.append(f"  {self.__get_input_type_name(input_1)} {self._get_name_1(is_obtained_vs_expected)}:"),
                msg_lines.append(self.__represent_input(input_1, 4))
            if input_2 is not None:
                msg_lines.append(f"  {self.__get_input_type_name(input_2)} {self._get_name_2(is_obtained_vs_expected)}:"),
                msg_lines.append(self.__represent_input(input_2, 4))

        return "\n".join(msg_lines)






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from behave.runner import Runner
from behave.parser import parse_feature
from holado.holado_config import Config


logger = logging.getLogger(__name__)



class IndependantRunner(Runner):
    """
    Independant Behave runner.
    
    WARNING: with current implementation, many files must exist in current working directory:
        - a folder "features" with at least one .feature file, even empty
        - a folder "steps" with a x_steps.py file importing all step files
        - a file "environment.py" containing "from XXX.behave_environment import *"
    """
    
    def __init__(self, config, step_paths=None):
        super().__init__(config)
        logger.info(f"Using behave runner IndependantRunner({step_paths=})")
        self.__step_paths = step_paths
        
    def load_step_definitions(self, extra_step_paths=None):
        # TODO: replace default implementation by loading steps in self.step_paths
        super().load_step_definitions(extra_step_paths)
        
        # logger.warning(f"++++ Loading step definitions ({self.__step_paths=})")
        # super().load_step_definitions(extra_step_paths=self.__step_paths)
        # logger.warning(f"++++ Loaded steps: {the_step_registry.steps}")
        
    def setup_paths(self):
        # TODO: replace default implementation by setting runtime defined paths passed in constructor
        super().setup_paths()
        
        # logger.warning(f"++++ setup_paths: {self.__step_paths=}")
        # self.path_manager.add(self.__step_paths)
        
    def run_model(self, features=None):
        # Use a custom feature that waits 7 days
        feature_text = \
            """
            Feature: Fake
                Scenario: Fake
                    When wait {runner_session_timeout} seconds
            """.format(runner_session_timeout=Config.session_timeout_seconds)
        feature = parse_feature(feature_text)
        
        return super().run_model([feature])






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tables.table_with_header import TableWithHeader
from builtins import object
from holado_core.common.tables.table import Table
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_core.common.tools.tools import Tools, reversed_range
from holado_python.standard_library.typing import Typing
import copy
from holado_value.common.tools.value_types import ValueTypes
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_value.common.tables.comparators.table_2_value_table_row_comparator import Table2ValueTable_RowComparator
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tables.comparators.table_comparator_manager import TableComparatorManager
from holado_core.common.tables.comparators.table_row_comparator import TableRowComparator
from holado_core.common.tables.comparators.table_comparator import TableComparator
from holado_core.common.tables.comparators.string_table_comparator import StringTableComparator
from holado_core.common.tables.table_row import TableRow
from holado_core.common.tables.comparators.string_table_row_comparator import StringTableRowComparator

logger = logging.getLogger(__name__)


class TableManager(object):
    
    @classmethod
    def is_table_with_header(cls, table):
        return isinstance(table, TableWithHeader)

    @classmethod
    def new_name_value_table(cls):
        res = TableWithHeader()
        res.header.add_cells_from_contents(["Name", "Value"])
        return res

    #TODO: move in TableConverter
    @classmethod
    def convert_object_attributes_2_name_value_table(cls, obj):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting object to name/value table")

        res = cls.new_name_value_table()

        # Set body
        attributes = Typing.get_object_attributes(obj)
        for attr in attributes:
            res.add_row(cells_content=attr)
            
        return res
    
    #TODO: move in TableConverter
    @classmethod
    def convert_list_2_column_table(cls, el_list):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting list to column table")

        res = Table()
        
        # Set body
        for el in el_list:
            res.add_row(cells_content=[el])
            
        return res
 
    #TODO: move in TableConverter
    @classmethod
    def convert_object_list_2_table_with_attributes_as_column(cls, list_obj):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting object list to table with object attributes as column...")
        res = TableWithHeader()

        for obj in list_obj:
            if not res.header:
                if isinstance(obj, dict):
                    header_names = list(obj.keys())
                else:
                    header_names = Typing.get_object_attribute_names(obj)
                res.header.add_cells_from_contents(cells_content=header_names)
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Result table - set header {res.header.represent(0)}")
            
            if isinstance(obj, dict):
                values_by_name = obj
            else:
                values_by_name = Typing.get_object_attribute_names(obj)
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Result table - add row with {values_by_name}")
            res.add_row(contents_by_colname=values_by_name)

        return res
 
    #TODO: move in TableConverter
    @classmethod
    def convert_object_list_2_column_table(cls, list_obj):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting object list to column table...")
        res = Table()

        for obj in list_obj:
#            logger.debug(f"+++++ add row with object: {obj}")
            res.add_row(cells_content=[obj])

        return res
        
    
    #TODO: move in TableConverter
    @classmethod
    def convert_dict_2_name_value_table(cls, obj):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting dictionary to name/value table...")
        res = cls.new_name_value_table()

        sorted_dict = dict(sorted(obj.items()))
        for entry in sorted_dict.items():
            res.add_row(cells_content=entry)
            
        return res
    
    #TODO: move in TableConverter
    @classmethod
    def convert_dict_2_table_with_keys_as_column(cls, obj):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting dictionary to table with keyx as column name...")
        res = TableWithHeader()
        
        res.add_row()
        sorted_dict = dict(sorted(obj.items()))
        for key, value in sorted_dict.items():
            res.add_column(name=key, cells_content=[value])
            
        return res
    
    @classmethod
    def compare_tables(cls, table_1, table_2, compare_params_table, reorder_columns_table_1=True, reorder_columns_table_2=True, raise_exception=True, do_log=True):
        compared_tables, columns_compare_methods, compared_columns, expected_table_columns = cls._extract_compared_data(table_1, table_2, compare_params_table, reorder_columns_table_1, reorder_columns_table_2)
        
        # Verify source table headers
        res = cls._verify_expected_table_columns(table_1, table_2, expected_table_columns, raise_exception=raise_exception)
        if not res:
            return res
        
        # Verify compared table headers
        res = cls._verify_compared_table_columns(compared_tables, compared_columns, raise_exception=raise_exception)
        if not res:
            return res
        
        # Compare table contents
        if columns_compare_methods is not None:
            cells_comparators = [TableComparatorManager.convert_compare_method_2_TableCellComparator(cm.value) if cm.value_type != ValueTypes.NotApplicable else None for cm in columns_compare_methods]
            row_comparator = TableRowComparator(cells_comparators=cells_comparators)
            comparator = TableComparator(row_comparator=row_comparator)
        else:
            comparator = StringTableComparator()
        try:
            res = comparator.equals(compared_tables[0], compared_tables[1], is_obtained_vs_expected=True, raise_exception=raise_exception, do_log=do_log)
        except FunctionalException as exc:
            raise FunctionalException(f"Tables are different (obtained = table 1 ; expected = table 2):\n{Tools.indent_string(4, exc.message)}") from exc
        
        return res
    
    @classmethod
    def _extract_compared_data(cls, table_1, table_2, compare_params_table, reorder_columns_table_1, reorder_columns_table_2):
        if not isinstance(compare_params_table, ValueTableWithHeader):
            raise TechnicalException(f"Compare parameters table is expected to be a ValueTableWithHeader (obtained type: {Typing.get_object_class_fullname(compare_params_table)})")
        
        res_t1 = copy.copy(table_1)
        res_t2 = copy.copy(table_2)
        colnames_t1 = compare_params_table.get_column(name="Column Names 1")
        colnames_t2 = compare_params_table.get_column(name="Column Names 2")
        has_compare_method = compare_params_table.has_column(name="Compare method", raise_exception=False)
        columns_compare_methods = compare_params_table.get_column(name="Compare method") if has_compare_method else None
        
        # Extract compared data
        compare_colnames_t1 = copy.copy(colnames_t1)
        compare_colnames_t2 = copy.copy(colnames_t2)
        exp_colnames_t1 = copy.copy(colnames_t1)
        exp_colnames_t2 = copy.copy(colnames_t2)
        for ind in reversed_range(len(colnames_t1)):
            if colnames_t1[ind].value_type == ValueTypes.NotApplicable or colnames_t2[ind].value_type == ValueTypes.NotApplicable:
                del compare_colnames_t1[ind]
                del compare_colnames_t2[ind]
                if has_compare_method:
                    del columns_compare_methods[ind]
            if colnames_t1[ind].value_type == ValueTypes.NotApplicable:
                cn_t2 = colnames_t2[ind].string_content
                res_t2.remove_column(name=cn_t2)
                del exp_colnames_t1[ind]
            if colnames_t2[ind].value_type == ValueTypes.NotApplicable:
                cn_t1 = colnames_t1[ind].string_content
                res_t1.remove_column(name=cn_t1)
                del exp_colnames_t2[ind]
        
        # Reorder tables
        if reorder_columns_table_1:
            table_1.order_columns(names=exp_colnames_t1.cells_value)
            res_t1.order_columns(names=compare_colnames_t1.cells_value)
        if reorder_columns_table_2:
            table_2.order_columns(names=exp_colnames_t2.cells_value)
            res_t2.order_columns(names=compare_colnames_t2.cells_value)
        
        return (res_t1, res_t2), columns_compare_methods, (compare_colnames_t1, compare_colnames_t2), (exp_colnames_t1, exp_colnames_t2)
    
    @classmethod
    def _verify_expected_table_columns(cls, table_1, table_2, expected_table_columns, raise_exception=True):
        comp = Table2ValueTable_RowComparator()
        try:
            comp.equals(table_1.header, expected_table_columns[0])
        except VerifyException as exc:
            if raise_exception:
                msg_list = ["Column Names 1 is not matching table 1 columns (columns order is mandatory):",
                            "    Table 1 columns: " + table_1.header.represent(0),
                            "    Column Names 1:  " + expected_table_columns[0].represent(0) ]
                raise FunctionalException("\n".join(msg_list) + f"\n{'-'*20}\n{exc}") from exc
            else:
                return False
        
        try:
            sorted_table_2_header = copy.copy(table_2.header)
            sorted_table_2_header.sort()
            sorted_exp_colnames_t2 = copy.copy(expected_table_columns[1])
            sorted_exp_colnames_t2.sort()
            comp.equals(sorted_table_2_header, sorted_exp_colnames_t2)
        except VerifyException as exc:
            if raise_exception:
                msg_list = ["Column Names 2 is not matching table 2 columns (columns order is not mandatory):",
                            "    Table 2 columns: " + table_2.header.represent(0),
                            "    Column Names 2:  " + expected_table_columns[1].represent(0) ]
                raise FunctionalException("\n".join(msg_list) + f"\n{'-'*20}\n{exc}") from exc
            else:
                return False
        
        return True
    
    @classmethod
    def _verify_compared_table_columns(cls, compared_tables, compared_columns, raise_exception=True):
        if compared_tables[0].nb_columns != len(compared_columns[0]):
            if raise_exception:
                msg_list = ["Compared columns are not matching table 1 columns:",
                            "    Table 1 columns:  " + compared_tables[0].header.represent(0),
                            "    compared columns: " + compared_columns[0].represent(0) ]
                raise FunctionalException("\n".join(msg_list))
            else:
                return False
        if compared_tables[1].nb_columns != len(compared_columns[1]):
            if raise_exception:
                msg_list = ["Compared columns are not matching table 2 columns:",
                            "    Table 2 columns:  " + compared_tables[1].header.represent(0),
                            "    compared columns: " + compared_columns[1].represent(0) ]
                raise FunctionalException("\n".join(msg_list))
            else:
                return False
        
        for ind in range(len(compared_columns[0])):
            index_t1 = compared_tables[0].get_column_index(name=compared_columns[0][ind].string_content)
            index_t2 = compared_tables[1].get_column_index(name=compared_columns[1][ind].string_content)
            if index_t1 != index_t2:
                if raise_exception:
                    msg_list = ["Compared columns have not same indexes in respective tables:",
                                "    Table header 1: " + compared_tables[0].header.represent(0),
                                "    Table header 2: " + compared_tables[1].header.represent(0) ]
                    raise FunctionalException("\n".join(msg_list))
                else:
                    return False
        
        return True
    
    @classmethod
    def verify_table_is_with_header(cls, table, raise_exception=True):
        # Compare headers
        if not cls.is_table_with_header(table):
            if raise_exception:
                raise TechnicalException(f"Table is expected with header (obtained type: {type(table)})")
            else:
                return False
        return True
    
    @classmethod
    def verify_table_is_value_table(cls, table, raise_exception=True):
        # Compare headers
        res = cls.verify_table_is_with_header(table, raise_exception=raise_exception)
        if not res:
            return False
        
        # Build expected header
        expected = TableRow(cells_content=["Value"])
        
        # Verify header content
        comparator = StringTableRowComparator()
        return comparator.equals(expected, table.header, raise_exception=raise_exception, do_log=False)
        
    @classmethod
    def verify_table_is_name_value_table(cls, table, raise_exception=True):
        # Compare headers
        res = cls.verify_table_is_with_header(table, raise_exception=raise_exception)
        if not res:
            return False
        
        # Build expected header
        expected = TableRow(cells_content=["Name", "Value"])
        
        # Verify header content
        comparator = StringTableRowComparator()
        return comparator.equals(expected, table.header, raise_exception=raise_exception, do_log=False)
        
    @classmethod
    def verify_table_is_x_name_value_table(cls, table, x_name, raise_exception=True):
        # Compare headers
        res = cls.verify_table_is_with_header(table, raise_exception=raise_exception)
        if not res:
            return False
        
        # Build expected header
        expected = TableRow(cells_content=[x_name, "Name", "Value"])
        
        # Verify header content
        comparator = StringTableRowComparator()
        return comparator.equals(expected, table.header, raise_exception=raise_exception, do_log=False)
        
    @classmethod
    def verify_table_is_x_table(cls, table, *x_names, raise_exception=True):
        # Compare headers
        res = cls.verify_table_is_with_header(table, raise_exception=raise_exception)
        if not res:
            return False
        
        # Build expected header
        expected = TableRow(cells_content=x_names)
        
        # Verify header content
        comparator = StringTableRowComparator()
        return comparator.equals(expected, table.header, raise_exception=raise_exception, do_log=False)
    
    @classmethod
    def represent_table(cls, table, indent=0):
        res_list = []
        
        if cls.is_table_with_header(table):
            res_list.append("| " + " | ".join(table.header.cells_content) + " |")
        
        for row in table.rows:
            res_list.append("| " + " | ".join(row.cells_content) + " |")
            
        return Tools.indent_string(indent, "\n".join(res_list))
    
    @classmethod
    def set_object_attributes_according_name_value_table(cls, obj, table):
        # Verify table structure
        cls.verify_table_is_name_value_table(table)
        
        for row in table.rows:
            setattr(obj, row.get_cell(0).content, row.get_cell(1).content)




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.common.handlers.object import Object
import abc
from holado.common.handlers.undefined import default_value, any_value

logger = logging.getLogger(__name__)


class DBAuditManager(Object):
    """Base class for audit managers.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, name, db_client):
        super().__init__(name)
        
        self.__db_client = db_client
    
    @property
    def _db_client(self):
        return self.__db_client
    
    def audit_table(self, table_name, operation_types=any_value, **kwargs):
        raise NotImplementedError()

    def _get_audit_table_name(self, audit_table_name=default_value):
        if audit_table_name is default_value:
            return "_audit"
        else:
            return audit_table_name

    def _get_audit_table_sql_create(self, audit_table_name=default_value):
        raise NotImplementedError()
    
    def ensure_audit_table_exists(self, audit_table_name=default_value):
        audit_table_name = self._get_audit_table_name(audit_table_name)
        create_sql = self._get_audit_table_sql_create(audit_table_name=audit_table_name)
        self._db_client.create_table(audit_table_name, create_sql, raise_if_exist=False, do_commit=True, do_audit=False)
    


class TriggerAuditManager(DBAuditManager):
    def __init__(self, name, db_client):
        super().__init__(name, db_client)
    
    def audit_table(self, table_name, operation_types=any_value, audit_table_name=default_value, **kwargs):
        self.ensure_audit_table_exists(audit_table_name)
        
        if operation_types is any_value:
            operation_types = ['insert', 'update', 'delete']
        else:
            operation_types = [ot.lower() for ot in operation_types]
        
        for op_type in operation_types:
            self.drop_trigger_of_audit_table_operation(table_name, op_type, audit_table_name, **kwargs)
            self.create_trigger_to_audit_table_operation(table_name, op_type, audit_table_name, **kwargs)
    
    def get_trigger_name(self, table_name, operation_type):
        return f"audit_{table_name}_on_{operation_type}"
    
    def drop_trigger_of_audit_table_operation(self, table_name, operation_type, audit_table_name, **kwargs):
        drop_sql = self._get_drop_trigger_sql_of_audit_table_operation(table_name, operation_type, audit_table_name)
        
        do_commit = kwargs.pop('do_commit', True)
        self._db_client.execute(drop_sql, do_commit=do_commit, **kwargs)
    
    def _get_drop_trigger_sql_of_audit_table_operation(self, table_name, operation_type):
        raise NotImplementedError()
    
    def create_trigger_to_audit_table_operation(self, table_name, operation_type, audit_table_name, **kwargs):
        create_sql = self._get_create_trigger_sql_to_audit_table_operation(table_name, operation_type, audit_table_name)
        
        do_commit = kwargs.pop('do_commit', True)
        self._db_client.execute(create_sql, do_commit=do_commit, **kwargs)
    
    def _get_create_trigger_sql_to_audit_table_operation(self, table_name, operation_type):
        raise NotImplementedError()







import holado


# Minimal HolAdo initialization for scripts
holado.initialize_for_script()

import logging
logging.print("After 'holado.initialize_for_script()', HolAdo is initialized minimally for scripts with:")
logging.print("  - add logging levels 'trace' and 'print'")
logging.print("  - do logs on console but not in file")
logging.print("  - a default session context is initialialized with all available HolAdo services")
logging.print("  - no session folder is created")


# Launch manually garbage collector periodically
from holado.common.tools.gc_manager import GcManager
from holado.common.handlers.undefined import default_value
GcManager.collect_periodically(collect_periodicity=default_value) 
logging.print("After 'GcManager.collect_periodically(collect_periodicity=default_value)', garbage collector is launched periodically (with default periodicity that is 10 s)")








#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.handlers.object import DeleteableObject
import os
from typing import AnyStr, List
from holado_python.standard_library.typing import Typing
from holado_core.common.exceptions.technical_exception import TechnicalException
import copy


class File(DeleteableObject):
    """
    Manage a file
    """

    def __init__(self, path, auto_flush=True, do_open=False, **open_kwargs):
        """ Manage a file
        @param path: file path
        @param do_open: if file must be open during this file initialization
        @param open_kwargs: arguments to open method
        """
        super().__init__(f"file '{path}'")

        self.__path = path
        self.__auto_flush = auto_flush
        self.__file = None
        self.__open_kwargs = open_kwargs
        if do_open:
            self.open()

    def _delete_object(self):
        self.close()
    
    def __enter__(self):
        if not self.is_open:
            self.open()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
    
    @property
    def path(self):
        return self.__path

    @property
    def internal_file(self):
        return self.__file
    
    @property
    def is_open(self):
        return self.__file is not None

    def open(self, **kwargs):
        if self.is_open:
            raise TechnicalException(f"File '{self.path}' is already opened")
        open_kwargs = copy.copy(self.__open_kwargs)
        if kwargs:
            open_kwargs.update(kwargs)
        
        self.__file = open(self.__path, **open_kwargs)

    def close(self):
        if self.__file:
            self.__file.close()
            self.__file = None

    def write(self, data):
        res = self.internal_file.write(data)
        if self.__auto_flush:
            self.internal_file.flush()
        return res

    def writelines(self, lines, add_line_sep=True):
        for line in lines:
            self.internal_file.write(line)
            if add_line_sep and not line.endswith(os.linesep):
                self.internal_file.write(os.linesep)
        if self.__auto_flush:
            self.internal_file.flush()

    def read(self, n: int = -1) -> AnyStr:
        return self.internal_file.read(n)

    def readline(self, limit: int = -1) -> AnyStr:
        return self.internal_file.readline(limit)

    def readlines(self, hint: int = -1, strip_newline=False) -> List[AnyStr]:
        res = self.internal_file.readlines(hint)
        if strip_newline:
            res = [l.strip('\n') for l in res]
        return res



    @classmethod
    def read_file_content(cls, path, **open_kwargs):
        with File(path, **open_kwargs) as file:
            res = file.read()
        return res

    @classmethod
    def read_file_content_in_base64(cls, path, **open_kwargs):
        import base64
        
        # Force to read file in binary mode
        open_kwargs['mode'] = 'rb'
        
        with File(path, **open_kwargs) as fin:
            content = fin.read()
        res = base64.b64encode(content)
        
        return res

    @classmethod
    def read_file_content_in_hexadecimal(cls, path, **open_kwargs):
        # Force to read file in binary mode
        open_kwargs['mode'] = 'rb'
        
        with File(path, **open_kwargs) as fin:
            content = fin.read()
        res = content.hex()
        
        return res

    @classmethod
    def write_file_with_content(cls, path, content):
        if isinstance(content, str):
            with File(path, mode='wt') as fout:
                fout.write(content)
        elif isinstance(content, bytes):
            with File(path, mode='wb') as fout:
                fout.write(content)
        else:
            raise TechnicalException(f"Unexpected content type {Typing.get_object_class_fullname(content)} (allowed types: str, bytes)")








#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from builtins import super
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_value.common.tables.comparators.table_2_value_table_cell_comparator import Table2ValueTable_CellComparator
from holado_ui_selenium.ui.gui.selenium.inspectors.selenium_inspector import SeleniumInspector
from holado_core.common.finders.tools.find_parameters import FindParameters

logger = logging.getLogger(__name__)


class Selenium2Value_TableCellComparator(Table2ValueTable_CellComparator):
    def __init__(self, inspector:SeleniumInspector=None):
        super().__init__()
        self.__inspector = inspector

    def _equals_symbol(self, cell_1, cell_symbol, is_obtained_vs_expected=True, raise_exception=True, do_log=True):
        res = True
        
        # Get cell selenium element
        cell_1_element = cell_1.content
        
        symbol_value = cell_symbol.value
        if symbol_value is None:
            if cell_1_element is not None:
                if raise_exception:
                    raise VerifyException("Cell value is not None (cell: [{}])".format(cell_1))
                else:
                    res = False
        else:
            # Find symbol element
            element = self.__inspector.get_finder_symbol(symbol_value).find_in(cell_1_element, FindParameters.default(raise_exception))
            
            res = (element is not None)
            if not res and raise_exception:
                raise FunctionalException(f"Unable to find symbol '{symbol_value}'.")
        
        return res
    
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_scripting.text.base.base_function import BaseFunction
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_core.common.tools.string_tools import StrTools
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class FunctionHexToBytes(BaseFunction):

    def __init__(self, expression_evaluator):
        self.__expression_evaluator = expression_evaluator
        
    def apply(self, args):
        # Verify arguments
        if not isinstance(args, list):
            raise TechnicalException("Arguments must be a list")
        if len(args) != 1:
            raise FunctionalException("Function 'HexToBytes' requires one argument.")

        src = args[0]
        
        # Evaluate parameter
        if isinstance(src, str):
            _, src_eval = self.__expression_evaluator.evaluate_expression(src, log_level=logging.ERROR)
            if isinstance(src_eval, bytes) or isinstance(src_eval, str):
                src = src_eval
            else:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Skipping evaluation: evaluation of expression '{src}' gives [{src_eval}] (type: {Typing.get_object_class_fullname(src_eval)}) that is not of type str or bytes")
        
        return StrTools.hex_to_bytes(src)
    



# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_protobuf.ipc.protobuf.protobuf_messages import ProtobufMessages
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.common.tools.datetime import DateTime
from holado_python.standard_library.typing import Typing
from holado_protobuf.ipc.protobuf.protobuf_converter import SortOrder

logger = logging.getLogger(__name__)


if ProtobufMessages.is_available():
    
    # TODO: check if needed
    import google.protobuf.pyext
    
    def __get_protobuf_converter():
        return SessionContext.instance().protobuf_converter
    
    def __get_protobuf_messages():
        return SessionContext.instance().protobuf_messages
    
    def __get_protobuf_modifier():
        return SessionContext.instance().protobuf_modifier
    
    def __get_scenario_context():
        return SessionContext.instance().get_scenario_context()
    
    def __get_text_interpreter():
        return __get_scenario_context().get_text_interpreter()
    
    def __get_variable_manager():
        return __get_scenario_context().get_variable_manager()
    
    def __get_session_context():
        return SessionContext.instance()
    
    @Step(r"(?P<var_name>{Variable}) = convert Protobuf object (?P<obj_var_name>{Variable}) to datetime")
    def step_impl(context, var_name, obj_var_name):  # @DuplicatedSignature
        from google.protobuf import timestamp_pb2
        
        var_name = StepTools.evaluate_variable_name(var_name)
        obj = StepTools.evaluate_variable_value(obj_var_name)
        
        if not isinstance(obj, timestamp_pb2.Timestamp):  # @UndefinedVariable
            raise TechnicalException(f"Object must be of type google.protobuf.Timestamp (obtained type: {Typing.get_object_class_fullname(obj)})")
        
        res = DateTime.seconds_nanos_to_datetime(obj.seconds, obj.nanos)
        
        __get_variable_manager().register_variable(var_name, res)

    @Step(r"(?P<var_name>{Variable}) = convert list (?P<list_var_name>{Variable}) to list of Protobuf field (?P<attribute_name>{Str})")
    def step_impl(context, var_name, list_var_name, attribute_name):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        attr_name = StepTools.evaluate_scenario_parameter(attribute_name)
        list_obj = StepTools.evaluate_variable_value(list_var_name)
        
        res = []
        for index, obj in enumerate(list_obj):
            if hasattr(obj, attr_name):
                attr_val = getattr(obj, attr_name)
                if ProtobufMessages.is_object_repeated(attr_val):
                    attr_val = [v for v in attr_val]
                # logger.debug(f"Result list - add field value [{attr_val}] (type: {Typing.get_object_class_fullname(attr_val)} ; dir: {dir(attr_val)})")
                res.append(attr_val)
            else:
                raise FunctionalException(f"In list, object of index {index} hasn't attribute '{attr_name}'")
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r"(?P<var_name>{Variable}) = convert list (?P<list_var_name>{Variable}) to merged list of Protobuf repeated field (?P<attribute_name>{Str})")
    def step_impl(context, var_name, list_var_name, attribute_name):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        attr_name = StepTools.evaluate_scenario_parameter(attribute_name)
        list_obj = StepTools.evaluate_variable_value(list_var_name)
        
        res = []
        for index, obj in enumerate(list_obj):
            if hasattr(obj, attr_name):
                attr_val = getattr(obj, attr_name)
                if not ProtobufMessages.is_object_repeated(attr_val):
                    raise FunctionalException(f"In list, object of index {index} has field '{attr_name}' but it isn't repeated")
                
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"In list, object of index {index} has field '{attr_name}' repeated {len(attr_val)} times")
                for val in attr_val:
                    # logger.debug(f"Result list - add field value [{val}] (type: {Typing.get_object_class_fullname(val)} ; dir: {dir(val)})")
                    res.append(val)
            else:
                raise FunctionalException(f"In list, object of index {index} hasn't attribute '{attr_name}'")
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?P<var_name>{Variable}) = convert list (?P<list_var_name>{Variable}) to list of (?P<proto_type_str>{Str}) Protobuf objects")
    def step_impl(context, var_name, list_var_name, proto_type_str):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        list_obj = StepTools.evaluate_variable_value(list_var_name)
        proto_type_str = StepTools.evaluate_scenario_parameter(proto_type_str)
        
        res = []
        for index, obj in enumerate(list_obj):
            new_obj = __get_protobuf_messages().new_object(proto_type_str)
            field_names = __get_protobuf_messages().get_object_field_names(obj, recursive=True, uncollapse_repeated=True, add_repeated_index=True, with_unset=False)
            for fn in field_names:
                try:
                    value = __get_protobuf_messages().get_object_field_value(obj, fn)
                    __get_protobuf_messages().set_object_field_value(new_obj, fn, value)
                except Exception as exc:
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"For object of index {index}, failed to set field '{fn}': {str(exc)}")
            res.append(new_obj)
        
        __get_variable_manager().register_variable(var_name, res)
    
    
    @Given(r'(?P<var_name>{Variable}) = convert list (?P<list_obj_str>{Variable}) to table with Protobuf fields as columns')
    def step_impl(context, var_name, list_obj_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        list_obj = StepTools.evaluate_variable_value(list_obj_str)
        
        res = __get_protobuf_converter().create_table_with_protobuf_fields_as_columns(list_obj, recursive=False)
    
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r'(?P<var_name>{Variable}) = convert list (?P<list_obj_str>{Variable}) to table with Protobuf fields as columns recursively')
    def step_impl(context, var_name, list_obj_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        list_obj = StepTools.evaluate_variable_value(list_obj_str)
        
        res = __get_protobuf_converter().create_table_with_protobuf_fields_as_columns(list_obj, recursive=True)
    
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r'(?P<var_name>{Variable}) = convert list (?P<list_obj_str>{Variable}) to table with Protobuf fields as columns recursively and repeated fields uncollapsed')
    def step_impl(context, var_name, list_obj_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        list_obj = StepTools.evaluate_variable_value(list_obj_str)
        
        res = __get_protobuf_converter().create_table_with_protobuf_fields_as_columns(list_obj, recursive=True, uncollapse_repeated=True)
    
        __get_variable_manager().register_variable(var_name, res)
    
     
    @Step(r"(?P<var_name>{Variable}) = new Protobuf object of type (?P<proto_type_str>{Str})")
    def step_impl(context, var_name, proto_type_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        proto_type_str = StepTools.evaluate_scenario_parameter(proto_type_str)
        table = BehaveStepTools.get_step_table(context)
        
        res = __get_protobuf_messages().new_message(proto_type_str, fields_table=table)
            
        __get_variable_manager().register_variable(var_name, res)  

    @Step(r"(?P<var_name>{Variable}) = serialize Protobuf object (?P<proto_str>{Variable})")
    def step_impl(context, var_name, proto_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        obj = StepTools.evaluate_variable_value(proto_str)
        res = obj.SerializeToString()
        __get_variable_manager().register_variable(var_name, res) 
       
    @Step(r"(?P<var_name>{Variable}) = serialize list (?P<obj_list>{Variable}) of Protobuf objects")
    def step_impl(context, var_name, obj_list):
        var_name = StepTools.evaluate_variable_name(var_name)
        obj_list = StepTools.evaluate_variable_value(obj_list)
        
        res = []
        for obj in obj_list:
            obj = obj.SerializeToString()
            res.append(obj)
        
        __get_variable_manager().register_variable(var_name, res)
       
    @Step(r"(?P<var_name>{Variable}) = unserialize string (?P<ser_str>{Str}) as (?P<proto_type_str>{Str}) Protobuf object")
    def step_impl(context, var_name, ser_str, proto_type_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        ser_str = StepTools.evaluate_scenario_parameter(ser_str)
        proto_type_str = StepTools.evaluate_scenario_parameter(proto_type_str)
    
        obj = __get_protobuf_messages().new_message(proto_type_str, serialized_string=ser_str)

        __get_variable_manager().register_variable(var_name, obj)
       
    @Step(r"(?P<var_name>{Variable}) = unserialize string list (?P<ser_list>{Variable}) as (?P<proto_type_str>{Str}) Protobuf objects")
    def step_impl(context, var_name, ser_list, proto_type_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        ser_list = StepTools.evaluate_variable_value(ser_list)
        proto_type_str = StepTools.evaluate_scenario_parameter(proto_type_str)
        
        res = []
        for ser_str in ser_list:
            obj = __get_protobuf_messages().new_message(proto_type_str, serialized_string=ser_str)
            res.append(obj)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r"(?P<var_name>{Variable}) = convert Protobuf object (?P<obj_var_name>{Variable}) to json object")
    def step_impl(context, var_name, obj_var_name):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        obj = StepTools.evaluate_variable_value(obj_var_name)
        res = __get_protobuf_converter().convert_protobuf_object_to_json_object(obj)
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r"(?P<var_name>{Variable}) = convert list (?P<obj_list_var_name>{Variable}) of Protobuf objects to json objects")
    def step_impl(context, var_name, obj_list_var_name):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        obj_list = StepTools.evaluate_variable_value(obj_list_var_name)
        
        res = []
        for obj in obj_list:
            res_obj = __get_protobuf_converter().convert_protobuf_object_to_json_object(obj)
            res.append(res_obj)
            
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r"(?P<var_name>{Variable}) = convert Protobuf object (?P<obj_var_name>{Variable}) to name/value table(?: with(?:(?P<with_names_str> names)?(?: and)?(?P<with_repeated_str> repeated)? uncollapsed)?(?: and)?(?P<with_unset_str> unset fields)?)?(?: \((?P<sort_alphabetically>sort alphabetically)?\))?")
    def step_impl(context, var_name, obj_var_name, with_names_str, with_repeated_str, with_unset_str, sort_alphabetically):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        obj = StepTools.evaluate_variable_value(obj_var_name)
        recursive = with_names_str is not None
        uncollapse_repeated = with_repeated_str is not None
        with_unset = with_unset_str is not None
        sort_order = SortOrder.Alphabetic if sort_alphabetically is not None else None
        
        res = __get_protobuf_converter().convert_protobuf_object_to_name_value_table(obj, recursive=recursive, uncollapse_repeated=uncollapse_repeated, with_unset=with_unset, sort_order=sort_order)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r"(?P<var_name>{Variable}) = value of Protobuf enum (?P<enum_fullname>{Str})")
    def step_impl(context, var_name, enum_fullname):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        enum_fullname = StepTools.evaluate_scenario_parameter(enum_fullname)
        res = __get_protobuf_messages().get_enum_value(fullname=enum_fullname)
        __get_variable_manager().register_variable(var_name, res)

    @Given(r"(?P<var_name>{Variable}) = name of value (?P<enum_value>{Str}) of Protobuf enum type (?P<enum_type_fullname>{Str})")
    def step_impl(context, var_name, enum_value, enum_type_fullname):  # @DuplicatedSignature
        var_name = StepTools.evaluate_variable_name(var_name)
        enum_value = StepTools.evaluate_scenario_parameter(enum_value)
        enum_type_fullname = StepTools.evaluate_scenario_parameter(enum_type_fullname)
        res = __get_protobuf_messages().get_enum_name(enum_value=enum_value, enum_type_fullname=enum_type_fullname)
        __get_variable_manager().register_variable(var_name, res)

    @Given(r"import Protobuf enum type (?P<enum_type_fullname>{Str})")
    def step_impl(context, enum_type_fullname):  # @DuplicatedSignature
        enum_type_fullname = StepTools.evaluate_scenario_parameter(enum_type_fullname)
        res = __get_protobuf_messages().get_enum_type(enum_type_fullname=enum_type_fullname)
        __get_variable_manager().register_variable(res.DESCRIPTOR.name, res)
        
        # Register each enum name fullname
        # Note: accept_expression=False is needed, otherwise, when importing a second enum with same package name, 
        #       previous line "enum_type_fullname = StepTools.evaluate_scenario_parameter(enum_type_fullname)" will fail
        for name, value in res.DESCRIPTOR.values_by_name.items():
            __get_variable_manager().register_variable(f"{res.DESCRIPTOR.full_name}.{name}", value.number, accept_expression=False)

    @Given(r"import values of Protobuf enum type (?P<enum_type_fullname>{Str})(?: with prefix (?P<prefix>{Str}))?")
    def step_impl(context, enum_type_fullname, prefix):  # @DuplicatedSignature
        enum_type_fullname = StepTools.evaluate_scenario_parameter(enum_type_fullname)
        prefix = StepTools.evaluate_scenario_parameter(prefix)
        et = __get_protobuf_messages().get_enum_type(enum_type_fullname=enum_type_fullname)
        for name, value in et.DESCRIPTOR.values_by_name.items():
            if prefix:
                __get_variable_manager().register_variable(prefix+name, value.number, accept_expression=False)
            else:
                __get_variable_manager().register_variable(name, value.number)
       
    @Step(r"(?P<var_name>{Variable}) = list (?P<obj_list>{Variable}) of Protobuf objects with timestamps shifted by (?P<shift_seconds>{Int}) seconds")
    def step_impl(context, var_name, obj_list, shift_seconds):
        var_name = StepTools.evaluate_variable_name(var_name)
        obj_list = StepTools.evaluate_variable_value(obj_list)
        shift_seconds = StepTools.evaluate_scenario_parameter(shift_seconds)
        
        res = __get_protobuf_modifier().shift_all_timestamps(obj_list, shift_seconds)
        
        __get_variable_manager().register_variable(var_name, res)







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from builtins import super
from holado_core.common.tables.table_cell import TableCell
from holado_core.common.tools.tools import Tools
import copy
from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class TableRow(list):
    def __init__(self, cells=None, cells_content=None, **kwargs):
        super().__init__()
        
        self.__content = None
        self.__parent = None
        
        if cells is not None:
            self.cells.extend(cells)
        elif cells_content is not None:
            self.add_cells_from_contents(cells_content, **kwargs)
        
    @property
    def content(self):
        """ Content of whole table row, relevant for some table types. """
        return self.__content
    
    @content.setter
    def content(self, content):
        """ Set content of whole table row, relevant for some table types. """
        self.__content = content
    
    @property
    def parent(self):
        """ Parent, relevant for some table types. It's usually the origin of its creation. """
        return self.__parent
    
    @parent.setter
    def parent(self, parent):
        """ Set parent. """
        self.__parent = parent
    
    @property
    def cells(self): 
        return self
    
    @property
    def cells_content(self): 
        return tuple(c.content for c in self.cells)
    
    @property
    def nb_cells(self): 
        return len(self.cells)
    
    def __copy__(self):
        cls = self.__class__
        res = cls()
        
        # Copy cells
        for cell in self.cells:
            res.add_cell(cell=copy.copy(cell))
            
        return res

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result
    
    def __repr__(self)->str:
        if self.nb_cells < 100:
            return f"<{self.__class__.__name__}>({self.nb_cells})[{self.represent()}]"
        else:
            return f"<{self.__class__.__name__}>({self.nb_cells})"

    def add_cell(self, cell=None, content=None, **kwargs):
        """Add a cell
        @param cell: Cell instance
        @param content: Cell content
        @param kwargs: Additional arguments needed by sub-classes.
        """
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"add_cell({cell=}, {content=} [{Typing.get_object_class_fullname(content)}])")
        if cell is not None:
            self.cells.append(cell)
        else:
            self.add_cell(cell=self._new_cell(content, **kwargs), **kwargs)
    
    def add_cells_from_contents(self, cells_content, **kwargs):
        """Add cells from their contents
        @param cells_content: Cells content
        @param kwargs: Additional arguments needed by sub-classes.
        """
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"add_cells_from_contents({cells_content=})")
        for cell_content in cells_content:
            self.add_cell(content=cell_content, **kwargs)
    
    def get_cell(self, index, **kwargs) -> TableCell :  # @UnusedVariable
        return self.cells[index]
    
    def _new_cell(self, content, **kwargs):  # @UnusedVariable
        return TableCell(cell_content=content)
    
    def remove_cell(self, index, **kwargs):  # @UnusedVariable
        return self.cells.pop(index)
    
    def keep_cells(self, indexes, **kwargs):
        for i in reversed(range(self.nb_cells)):
            if i not in indexes:
                self.remove_cell(i, **kwargs)
    
    def represent(self, indent = 0, value_prefix = None, value_postfix = None, marge_left = " ", marge_right = " "):
        res_list = []
        
        res_list.append(Tools.get_indent_string(indent))
        res_list.append("|")
        for cell in self.cells:
            if marge_left is not None:
                res_list.append(marge_left)
            
            if cell is not None:
                if value_prefix is not None or value_postfix is not None:
                    res_list.append(cell.represent(0, value_prefix, value_postfix))
                else:
                    res_list.append(cell.represent(0))
            else:
                res_list.append("{CELL_IS_NONE}")
            
            if marge_right is not None:
                res_list.append(marge_right)
            
            res_list.append("|")
        
        return "".join(res_list)

    def order_cells(self, indexes=None, **kwargs):
        orig_indexes = list(range(self.nb_cells))
        for cell_index, orig_cell_ind in enumerate(indexes):
            cur_cell_ind = orig_indexes.index(orig_cell_ind)
            if cur_cell_ind != cell_index:
                self.switch_cells(index_1=cell_index, index_2=cur_cell_ind, **kwargs)
                orig_indexes[cell_index], orig_indexes[cur_cell_ind] = orig_indexes[cur_cell_ind], orig_indexes[cell_index]
        
    def switch_cells(self, index_1, index_2, **kwargs):
        if index_1 == index_2:
            raise TechnicalException(f"The two indexes are equal (index = {index_1})")
    
        self.cells[index_1], self.cells[index_2] = self.cells[index_2], self.cells[index_1]




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.behave.behave import *  # @UnusedWildImport
from holado.common.context.session_context import SessionContext
from holado_test.scenario.step_tools import StepTools
import logging
import queue
from typing import Iterable
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_system.system.global_system import GlobalSystem
from holado_system.system.filesystem.file import File
from holado_python.standard_library.queue import IterableQueue,\
    IterableLifoQueue
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()



@Given(r"(?P<var_name>{Variable}) = new FIFO queue(?: of size (?P<size>{Int}))?(?: \((?:block on get: (?P<block_on_get>{Boolean}))?(?: ; )?(?:block on get timeout: (?P<block_on_get_timeout>{Float} s))?(?: ; )?(?:block on put: (?P<block_on_put>{Boolean}))?(?: ; )?(?:block on put timeout: (?P<block_on_put_timeout>{Float}) s)?\))?")
def step_impl(context, var_name, size, block_on_get, block_on_get_timeout, block_on_put, block_on_put_timeout):
    var_name = StepTools.evaluate_variable_name(var_name)
    size = StepTools.evaluate_scenario_parameter(size)
    block_on_get = StepTools.evaluate_scenario_parameter(block_on_get)
    block_on_get_timeout = StepTools.evaluate_scenario_parameter(block_on_get_timeout)
    block_on_put = StepTools.evaluate_scenario_parameter(block_on_put)
    block_on_put_timeout = StepTools.evaluate_scenario_parameter(block_on_put_timeout)
    
    res = IterableQueue(size, block_on_get, block_on_get_timeout, block_on_put, block_on_put_timeout)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = new LIFO queue(?: of size (?P<size>{Int}))?(?: \((?:block on get: (?P<block_on_get>{Boolean}))?(?: ; )?(?:block on get timeout: (?P<block_on_get_timeout>{Float} s))?(?: ; )?(?:block on put: (?P<block_on_put>{Boolean}))?(?: ; )?(?:block on put timeout: (?P<block_on_put_timeout>{Float}) s)?\))?")
def step_impl(context, var_name, size, block_on_get, block_on_get_timeout, block_on_put, block_on_put_timeout):
    var_name = StepTools.evaluate_variable_name(var_name)
    size = StepTools.evaluate_scenario_parameter(size)
    block_on_get = StepTools.evaluate_scenario_parameter(block_on_get)
    block_on_get_timeout = StepTools.evaluate_scenario_parameter(block_on_get_timeout)
    block_on_put = StepTools.evaluate_scenario_parameter(block_on_put)
    block_on_put_timeout = StepTools.evaluate_scenario_parameter(block_on_put_timeout)
        
    res = IterableLifoQueue(size, block_on_get, block_on_get_timeout, block_on_put, block_on_put_timeout)
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"put (?!all )(?P<value>{Any}) \(Queue: (?P<var_queue>{Variable})\)")
def step_impl(context, value, var_queue):
    value = StepTools.evaluate_scenario_parameter(value)
    q = StepTools.evaluate_variable_value(var_queue)
    
    q.put(value)

@Step(r"(?P<var_name>{Variable}) = get \(Queue: (?P<var_queue>{Variable})\)")
def step_impl(context, var_name, var_queue):
    var_name = StepTools.evaluate_variable_name(var_name)
    q = StepTools.evaluate_variable_value(var_queue)
    
    res = q.get()
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"task is done \(Queue: (?P<var_queue>{Variable})\)")
def step_impl(context, var_queue):
    q = StepTools.evaluate_variable_value(var_queue)
    
    q.task_done()

@Step(r"join \(Queue: (?P<var_queue>{Variable})\)")
def step_impl(context, var_queue):
    q = StepTools.evaluate_variable_value(var_queue)
    
    q.join()

@Step(r"(?P<var_name>{Variable}) = size \(Queue: (?P<var_queue>{Variable})\)")
def step_impl(context, var_name, var_queue):
    var_name = StepTools.evaluate_variable_name(var_name)
    q = StepTools.evaluate_variable_value(var_queue)
        
    res = q.qsize()
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"close queue \(Queue: (?P<var_queue>{Variable})\)")
def step_impl(context, var_queue):
    q = StepTools.evaluate_variable_value(var_queue)
    q.close()



@Step(r"put all of (?P<iterable_str>{Variable}) \(Queue: (?P<var_queue>{Variable})(?: ; nb of repeat: (?P<repeat_nb>{Int}))?(?: ; rate log period: (?P<rate_period_s>{Int}) s)?\)")
def step_impl(context, iterable_str, var_queue, repeat_nb, rate_period_s):
    iterable = StepTools.evaluate_scenario_parameter(iterable_str)
    it_name = StepTools.evaluate_variable_name(iterable_str)
    q = StepTools.evaluate_variable_value(var_queue)
    q_name = StepTools.evaluate_variable_name(var_queue)
    repeat_nb = StepTools.evaluate_scenario_parameter(repeat_nb)
    rate_period_s = StepTools.evaluate_scenario_parameter(rate_period_s)
    
    if not isinstance(iterable, Iterable):
        raise FunctionalException(f"Parameter [{iterable_str}] (type: {Typing.get_object_class_fullname(iterable)}) is not iterable")

    if repeat_nb is None:
        repeat_nb = 1
    nb_elements = len(iterable)
    
    beg = Tools.timer_s()
    s_end, t_end = 0, beg
    for num in range(1, repeat_nb+1):
        s_beg, t_beg = is_beg, it_beg = s_end, t_end
        ind_beg = 0
        
        for ind, value in enumerate(iterable):
            q.put(value)
            
            if rate_period_s and ind > 0 and ind % 10 == 0:
                it_end = Tools.timer_s()
                if it_end > it_beg + rate_period_s:
                    is_end = q.qsize()
                    rate = (is_beg + ind - ind_beg - is_end) / (it_end - it_beg)
                    logger.print(f"[{it_name} -> Queue {q_name} ; Iteration {num}] Current rate: {round(rate,2)} item/s ; Queue size: {is_end}")
                    ind_beg, is_beg, it_beg = ind, is_end, it_end
                    
            # Yield processor in case of multithreading
            if ind % 10 == 0:
                GlobalSystem.yield_processor()
            
        if rate_period_s:
            s_end, t_end = q.qsize(), Tools.timer_s()
            rate = (s_beg + nb_elements - s_end) / (t_end - t_beg)
            logger.print(f"[{it_name} -> Queue {q_name} ; Iteration {num}] Mean rate: {round(rate,2)} item/s ; Queue size: {s_end}")
        
    if rate_period_s:
        end, s_end = Tools.timer_s(), q.qsize()
        rate = (nb_elements * repeat_nb - s_end) / (end - beg)
        logger.print(f"[{it_name} -> Queue {q_name}] Mean rate: {round(rate,2)} item/s")

@Step(r"put all of (?P<iterable_str>{Variable}) by batch of (?P<batch_size>{Int}) items \(Queue: (?P<var_queue>{Variable})(?: ; nb of repeat: (?P<repeat_nb>{Int}))?(?: ; rate log period: (?P<rate_period_s>{Int}) s)?\)")
def step_impl(context, iterable_str, batch_size, var_queue, repeat_nb, rate_period_s):
    iterable = StepTools.evaluate_scenario_parameter(iterable_str)
    it_name = StepTools.evaluate_variable_name(iterable_str)
    batch_size = StepTools.evaluate_scenario_parameter(batch_size)
    q = StepTools.evaluate_variable_value(var_queue)
    q_name = StepTools.evaluate_variable_name(var_queue)
    repeat_nb = StepTools.evaluate_scenario_parameter(repeat_nb)
    rate_period_s = StepTools.evaluate_scenario_parameter(rate_period_s)
    
    if not isinstance(iterable, Iterable):
        raise FunctionalException(f"Parameter [{iterable_str}] (type: {Typing.get_object_class_fullname(iterable)}) is not iterable")

    if repeat_nb is None:
        repeat_nb = 1
    nb_elements = len(iterable)
    
    beg = Tools.timer_s()
    s_end, t_end = 0, beg
    for num in range(1, repeat_nb+1):
        s_beg, t_beg = is_beg, it_beg = s_end, t_end
        batch_counter_beg = batch_counter = 0
        batch = []
        
        for value in iterable:
            batch.append(value)
            if len(batch) >= batch_size:
                batch_counter += 1
                q.put(batch)
                batch = []
            
                if rate_period_s and batch_counter > 0 and batch_counter % 10 == 0:
                    it_end = Tools.timer_s()
                    if it_end > it_beg + rate_period_s:
                        is_end = q.qsize()
                        rate = (is_beg + batch_counter - batch_counter_beg - is_end) / (it_end - it_beg)
                        logger.print(f"[{it_name} -> Queue {q_name} ; Iteration {num}] Current rate: {round(rate,2)} item/s ; Queue size: {is_end}")
                        batch_counter_beg, is_beg, it_beg = batch_counter, is_end, it_end
                        
                # Yield processor in case of multithreading
                if batch_counter % 10 == 0:
                    GlobalSystem.yield_processor()
            
        if rate_period_s:
            s_end, t_end = q.qsize(), Tools.timer_s()
            rate = (s_beg + nb_elements - s_end) / (t_end - t_beg)
            logger.print(f"[{it_name} -> Queue {q_name} ; Iteration {num}] Mean rate: {round(rate,2)} item/s ; Queue size: {s_end}")
        
    if rate_period_s:
        end, s_end = Tools.timer_s(), q.qsize()
        rate = (nb_elements * repeat_nb - s_end) / (end - beg)
        logger.print(f"[{it_name} -> Queue {q_name}] Mean rate: {round(rate,2)} item/s")

@Step(r"put all lines of file (?P<file_str>{Variable})(?: \(offset: (?P<offset>{Int}) ; size: (?P<size>{Int})\))? \(Queue: (?P<var_queue>{Variable})(?: ; rate log period: (?P<rate_period_s>{Int}) s)?\)")
def step_impl(context, file_str, offset, size, var_queue, rate_period_s):
    file = StepTools.evaluate_scenario_parameter(file_str)
    file_name = StepTools.evaluate_variable_name(file_str)
    q = StepTools.evaluate_variable_value(var_queue)
    q_name = StepTools.evaluate_variable_name(var_queue)
    offset = StepTools.evaluate_scenario_parameter(offset)
    size = StepTools.evaluate_scenario_parameter(size)
    rate_period_s = StepTools.evaluate_scenario_parameter(rate_period_s)
    
    if not isinstance(file, File):
        raise FunctionalException(f"Parameter [{file_str}] is not a file object")

    beg = Tools.timer_s()
    s_beg, t_beg = s_end, t_end = 0, beg
    ind_beg = 0
    
    for ind, line in enumerate(file.internal_file):
        if ind < offset:
            continue
        elif size > 0 and ind >= offset + size:
            break
        
        q.put(line.strip('\n'))
        
        if rate_period_s and ind > 0 and ind % 10 == 0:
            t_end = Tools.timer_s()
            if t_end > t_beg + rate_period_s:
                s_end = q.qsize()
                rate = (s_beg + ind - ind_beg - s_end) / (t_end - t_beg)
                logger.print(f"[File {file_name} -> Queue {q_name}] Current rate: {round(rate,2)} item/s ; Queue size: {s_end}")
                ind_beg, s_beg, t_beg = ind, s_end, t_end
                
        # Yield processor in case of multithreading
        if ind > 0 and ind % 10 == 0:
            GlobalSystem.yield_processor()
        
    if rate_period_s:
        end, s_end = Tools.timer_s(), q.qsize()
        rate = (ind + 1 - s_end) / (end - beg)
        logger.print(f"[File {file_name} -> Queue {q_name}] Mean rate: {round(rate,2)} item/s")



@Step(r"for (?P<item_var_name>{VariableName}) in (?P<iterable_str>{Variable}), call steps, and put (?P<result_var_name>{VariableName}) \(Queue: (?P<var_queue>{Variable})(?: ; rate log period: (?P<rate_period_s>{Int}) s)?(?: ; format with: (?P<format_with>{Variable}(?:, {Variable})*))?\)")
def step_impl(context, item_var_name, iterable_str, result_var_name, var_queue, rate_period_s, format_with):
    item_var_name = StepTools.evaluate_variable_name(item_var_name)
    iterable = StepTools.evaluate_variable_value(iterable_str)
    it_name = StepTools.evaluate_variable_name(iterable_str)
    result_var_name = StepTools.evaluate_variable_name(result_var_name)
    q = StepTools.evaluate_variable_value(var_queue)
    q_name = StepTools.evaluate_variable_name(var_queue)
    rate_period_s = StepTools.evaluate_scenario_parameter(rate_period_s)
    format_with = StepTools.evaluate_variable_name(format_with)
    steps = BehaveStepTools.get_step_multiline_text(context, eval_params=EvaluateParameters.nothing())
    
    if format_with is not None:
        steps = StepTools.format_steps_with(steps, format_with.split(', '))
    
    beg = Tools.timer_s()
    t_beg = t_end = beg
    ind_beg = 0
    
    for ind, item in enumerate(iterable):
        __get_variable_manager().register_variable(item_var_name, item)
        execute_steps(steps)
        result = __get_variable_manager().get_variable_value(result_var_name)
        
        q.put(result)
        
        if rate_period_s and ind > 0 and ind % 10 == 0:
            t_end = Tools.timer_s()
            if t_end > t_beg + rate_period_s:
                s_end = q.qsize()
                rate = (ind - ind_beg) / (t_end - t_beg)
                logger.print(f"[{it_name} -> Queue {q_name}] Current rate: {round(rate,2)} item/s ; Queue size: {s_end}")
                ind_beg, t_beg = ind, t_end
                
        # Yield processor in case of multithreading
        if ind > 0 and ind % 10 == 0:
            GlobalSystem.yield_processor()
        
    if rate_period_s:
        end, s_end = Tools.timer_s(), q.qsize()
        rate = (ind + 1) / (end - beg)
        logger.print(f"[{it_name} -> Queue {q_name}] Mean rate: {round(rate,2)} item/s ; Queue size: {s_end}")


@Step(r"for (?P<item_var_name>{VariableName}) in queue, call steps \(Queue: (?P<var_queue>{Variable})(?: ; rate log period: (?P<rate_period_s>{Int}) s)?(?: ; format with: (?P<format_with>{Variable}(?:, {Variable})*))?\)")
def step_impl(context, item_var_name, var_queue, rate_period_s, format_with):
    item_var_name = StepTools.evaluate_variable_name(item_var_name)
    q = StepTools.evaluate_variable_value(var_queue)
    q_name = StepTools.evaluate_variable_name(var_queue)
    rate_period_s = StepTools.evaluate_scenario_parameter(rate_period_s)
    format_with = StepTools.evaluate_variable_name(format_with)
    steps = BehaveStepTools.get_step_multiline_text(context, eval_params=EvaluateParameters.nothing())
    
    if format_with is not None:
        steps = StepTools.format_steps_with(steps, format_with.split(', '))
    
    counter = 0
    beg = Tools.timer_s()
    c_beg, t_beg = counter, beg
    log_counter = 0
    
    try:
        while True:
            # Get next item
            item = q.get()
            if q.is_sentinel(item):
                break
            counter += 1
            
            __get_variable_manager().register_variable(item_var_name, item)
            try:
                execute_steps(steps)
            except Exception as exc:
                logger.warning(f"[Queue {q_name}] Failed to execute steps with {counter}'th item: [{Typing.get_object_class_fullname(exc)}] {str(exc)}")
            finally:
                q.task_done()
            
            # Log rate if needed
            if rate_period_s and counter % 10 == 0:
                t_end = Tools.timer_s()
                if t_end > t_beg + rate_period_s:
                    c_end, s_end = counter, q.qsize()
                    rate = (c_end - c_beg) / (t_end - t_beg)
                    log_counter += 1
                    logger.print(f"[Queue {q_name}] Rate: {int(rate)} msg/s ; Nb items: {counter} ; Queue size: {s_end}")
                    c_beg, t_beg = c_end, t_end
                    
            # Yield processor in case of multithreading
            if counter % 10 == 0:
                GlobalSystem.yield_processor()
    except queue.Empty:
        # Without block, or with timeout, this exception occurs when queue is empty
        pass
    
    if rate_period_s:
        end = Tools.timer_s()
        rate = counter / (end - beg)
        logger.print(f"[Queue {q_name}] Mean rate: {int(rate)} msg/s ; Nb messages: {counter}")








#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
import os
from holado_system.system.command.command import Command, CommandStates
from holado.common.context.session_context import SessionContext
import json
from holado_system.system.filesystem.file import File

logger = logging.getLogger(__name__)


class TcpbinManager(object):
    """
    Helper for tcpbin.com usage.
    """
    
    @classmethod
    def __get_path_manager(cls):
        return SessionContext.instance().path_manager
    
    @classmethod
    def get_client_certificates(cls):
        base_path = cls.__get_path_manager().get_local_resources_path('certificates')
        certfile_path = os.path.join(base_path, 'tcpbin.crt')
        keyfile_path = os.path.join(base_path, 'tcpbin.key')
        
        cls.ensure_tcpbin_certificates_are_valid(certfile_path, keyfile_path)
        
        return (certfile_path, keyfile_path)
    
    @classmethod
    def get_CAs(cls):
        base_path = cls.__get_path_manager().get_local_resources_path('certificates')
        ca_certfile_path = os.path.join(base_path, 'rootCACert.pem')
        ca_keyfile_path = os.path.join(base_path, 'rootCAKey.pem')
        
        cls.extract_tcpbin_ca_files(ca_certfile_path, ca_keyfile_path)
        
        return (ca_certfile_path, ca_keyfile_path)
    
    @classmethod
    def ensure_tcpbin_certificates_are_valid(cls, public_key_path, private_key_path, duration_seconds=600):
        cls.__get_path_manager().makedirs(public_key_path)
        cls.__get_path_manager().makedirs(private_key_path)
        
        # Verify if certificate has expired
        do_generate_certificates = True
        if os.path.exists(public_key_path):
            cmd = f"openssl x509 -checkend {duration_seconds} -noout -in '{public_key_path}'"
            command = Command(cmd, do_log_output=True, do_raise_on_stderr=False, executable="/bin/bash")
            command.start()
            command.join()
            if command.state is CommandStates.Success:
                do_generate_certificates = False
            elif command.return_code == 1:
                do_generate_certificates = True
            else:
                raise TechnicalException(f"Error while executing openssl command [{cmd}]: error code={command.return_code} ; stdout: [{command.stdout}] ; stderr: [{command.stderr}]")
        
        # Generate new certificates if needed
        if do_generate_certificates:
            cmd = f"curl -s https://tcpbin.com/api/client-cert"
            command = Command(cmd, do_log_output=True, do_raise_on_stderr=False, executable="/bin/bash")
            command.start()
            command.join()
            if command.state is not CommandStates.Success:
                raise TechnicalException(f"Error while executing command [{cmd}] : [{command.stderr}]")
            
            data = json.loads(command.stdout)
            File.write_file_with_content(public_key_path, data['cert'])
            File.write_file_with_content(private_key_path, data['key'])
    
    @classmethod
    def extract_tcpbin_ca_files(cls, ca_certfile_path, ca_keyfile_path):
        if not cls.__get_path_manager().check_file_exists(ca_certfile_path, raise_exception=False):
            cmd = f"curl -s https://tcpbin.com/rootCACert.pem"
            command = Command(cmd, do_log_output=True, do_raise_on_stderr=True, executable="/bin/bash")
            command.start()
            command.join()
            File.write_file_with_content(ca_certfile_path, command.stdout)
        
        if not cls.__get_path_manager().check_file_exists(ca_keyfile_path, raise_exception=False):
            cmd = f"curl -s https://tcpbin.com/rootCAKey.pem"
            command = Command(cmd, do_log_output=True, do_raise_on_stderr=True, executable="/bin/bash")
            command.start()
            command.join()
            File.write_file_with_content(ca_keyfile_path, command.stdout)
        








#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from enum import Enum
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class CompareOperator(str, Enum):
    Different = ("!=", "Equal", "Different")
    Equal = ("==", "Different", "Equal")
    Inferior = ("<", "SuperiorOrEqual", "Superior")
    InferiorOrEqual = ("<=", "Superior", "SuperiorOrEqual")
    Superior = (">", "InferiorOrEqual", "Inferior")
    SuperiorOrEqual = (">=", "Inferior", "InferiorOrEqual")
    
    def __new__(cls, value, not_name, switch_name):
        obj = str.__new__(cls, [value])
        obj._value_ = value
        obj.__not_name = not_name
        obj.__switch_name = switch_name
        return obj
    
    @property
    def not_(self):
        return CompareOperator[self.__not_name]
    
    @property
    def switch(self):
        return CompareOperator[self.__switch_name]
    

class Comparator(object):
    """
    Mother class of comparator objects.
    """
    __metaclass__ = abc.ABCMeta	
    
    def __init__(self, type_description):
        self.__type_description = type_description
        self.__do_convert_input1 = True
        self.__do_convert_input2 = True

    @property
    def do_convert_input1(self):
        return self.__do_convert_input1
    
    @do_convert_input1.setter
    def do_convert_input1(self, value: bool):
        self.__do_convert_input1 = value

    @property
    def do_convert_input2(self):
        return self.__do_convert_input2
    
    @do_convert_input2.setter
    def do_convert_input2(self, value: bool):
        self.__do_convert_input2 = value
    
    def equals(self, obj_1, obj_2, is_obtained_vs_expected=True, raise_exception=True, do_log=True):
        """
        Compare by equality.
        """
        return self.compare(obj_1, CompareOperator.Equal, obj_2, is_obtained_vs_expected, raise_exception, do_log=do_log, redirect_to_equals=False)

    def compare(self, obj_1, operator: CompareOperator, obj_2, is_obtained_vs_expected=True, raise_exception=True, do_log=True, redirect_to_equals=True):
        """
        Compare by given CompareOperator.
        """
        obj_1, obj_2 = self._convert_inputs(obj_1, obj_2, is_obtained_vs_expected)
        return self._compare_by_operator(obj_1, operator, obj_2, is_obtained_vs_expected, raise_exception, do_log=do_log, redirect_to_equals=redirect_to_equals)
    
    def _convert_inputs(self, obj_1, obj_2, is_obtained_vs_expected):
        if self.do_convert_input1:
            obj_1 = self._convert_input1(obj_1, self._get_name_1(is_obtained_vs_expected))
        if self.do_convert_input2:
            obj_2 = self._convert_input2(obj_2, self._get_name_2(is_obtained_vs_expected))
        return [obj_1, obj_2]
    
    def _convert_input1(self, obj, name):
        return self._convert_input(obj, name)
    
    def _convert_input2(self, obj, name):
        return self._convert_input(obj, name)
    
    def _convert_input(self, obj, name):
        """
        Method applied to compared objects before comparison.
        To compare similar types that are not comparable directly, override this method to convert objects in comparable types
        """ 
        # By default no conversion is applied
        return obj
    
    def _compare_by_operator(self, obj_1, operator: CompareOperator, obj_2, is_obtained_vs_expected=True, raise_exception=True, do_log=True, redirect_to_equals=True):
        if redirect_to_equals and operator == CompareOperator.Equal:
            return self.equals(obj_1, obj_2, is_obtained_vs_expected, raise_exception, do_log=do_log)
    
        res = self._compare_result(obj_1, operator, obj_2)
        
        # logger.debug("[{}] obj_1 equals obj_2 => {}\n    obj_1 = [{}]\n    obj_2 = [{}]".format(Typing.get_object_class(self).__name__, res, obj_1, obj_2))
        if not res and raise_exception:
            raise VerifyException(f"Match failure, {self.__type_description} values don't verify expression [{self._get_name_1(is_obtained_vs_expected)} {operator.value} {self._get_name_2(is_obtained_vs_expected)}]:\n  {self._get_name_1(is_obtained_vs_expected)}: {obj_1} (type: {Typing.get_object_class_fullname(obj_1)})\n  {self._get_name_2(is_obtained_vs_expected)}: {obj_2} (type: {Typing.get_object_class_fullname(obj_2)})")
        return res
    
    def _compare_result(self, obj_1, operator: CompareOperator, obj_2):
        if operator == CompareOperator.Different:
            return (obj_1 != obj_2)
        elif operator == CompareOperator.Equal:
            return (obj_1 == obj_2)
        elif operator == CompareOperator.Inferior:
            return (obj_1 < obj_2)
        elif operator == CompareOperator.InferiorOrEqual:
            return (obj_1 <= obj_2)
        elif operator == CompareOperator.Superior:
            return (obj_1 > obj_2)
        elif operator == CompareOperator.SuperiorOrEqual:
            return (obj_1 >= obj_2)
        else:
            raise TechnicalException(f"Unmanaged compare operator {operator}")
        
    def _get_name_1(self, is_obtained_vs_expected):
        if is_obtained_vs_expected == True:
            return "obtained"
        elif is_obtained_vs_expected == False:
            return "expected"
        else:
            return "1"
        
    def _get_name_2(self, is_obtained_vs_expected):
        if is_obtained_vs_expected == True:
            return "expected"
        elif is_obtained_vs_expected == False:
            return "obtained"
        else:
            return "2"




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_system.system.filesystem.file import File
import base64
import os.path
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing
from zipfile import ZipFile

logger = logging.getLogger(__name__)



def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()

def __get_report_manager():
    return SessionContext.instance().report_manager

def __get_path_manager():
    return SessionContext.instance().path_manager


@Then(r"file (?P<path>{Str}) (?P<exists_str>exists|doesn't exist)")
def step_impl(context, path, exists_str):  # @DuplicatedSignature
    path = StepTools.evaluate_scenario_parameter(path)
    exists = exists_str == "exists"
    
    __get_path_manager().check_file_exists(path, do_exist=exists, raise_exception=True)

@Given(r"(?P<var_name>{Variable}) = is path (?P<path>{Str}) an absolute file path")
def step_impl(context, var_name, path):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    path = StepTools.evaluate_scenario_parameter(path)
    res = os.path.isabs(path)
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = file name of path (?P<path>{Str})(?: \((?P<without_extension_str>without extension)\))?")
def step_impl(context, var_name, path, without_extension_str):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    path = StepTools.evaluate_scenario_parameter(path)
    
    res = os.path.basename(path)
    if without_extension_str is not None:
        res = os.path.splitext(res)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = open file (?P<path>{Str}) in mode (?P<mode>{Str})")
def step_impl(context, var_name, path, mode):
    var_name = StepTools.evaluate_variable_name(var_name)
    execute_steps(u"""
        Given {var_name} = open file {path}
            | Name    | Value  |
            | 'mode'  | {mode} |
        """.format(var_name=var_name, path=path, mode=mode))

@Given(r"(?P<var_name>{Variable}) = open file (?P<path>{Str})")
def step_impl(context, var_name, path):
    var_name = StepTools.evaluate_variable_name(var_name)
    path = StepTools.evaluate_scenario_parameter(path)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    open_kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
    
    __get_path_manager().makedirs(path)
    res = File(path, **open_kwargs)

    __get_variable_manager().register_variable(var_name, res)

@Given(r"close \(file: (?P<file>{Variable})\)")
def step_impl(context, file):
    file = StepTools.evaluate_scenario_parameter(file)
    file.close()

@When(r"write line (?P<line>{Str}) \(file: (?P<file>{Variable})\)")
def step_impl(context, line, file):
    line = StepTools.evaluate_scenario_parameter(line)
    file = StepTools.evaluate_scenario_parameter(file)
    
    file.writelines([line])
    
@When(r"write lines (?P<lines_list>{List}) \(file: (?P<file>{Variable})\)")
def step_impl(context, lines_list, file):
    lines_list = StepTools.evaluate_scenario_parameter(lines_list)
    file = StepTools.evaluate_scenario_parameter(file)
    
    file.writelines(lines_list)
    
@When(r"write (?P<data>{Str}) \(file: (?P<file>{Variable})\)")
def step_impl(context, data, file):
    data = StepTools.evaluate_scenario_parameter(data)
    file = StepTools.evaluate_scenario_parameter(file)
    
    file.write(data)

@Given(r"(?P<var_name>{Variable}) = content of(?:(?P<is_text_file_str> text))? file (?P<path>{Str})")
def step_impl(context, var_name, is_text_file_str, path):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    is_text_file = is_text_file_str is not None
    path = StepTools.evaluate_scenario_parameter(path)
    
    mode = 'rt' if is_text_file else 'rb'
    res = File.read_file_content(path, mode=mode)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = content of file (?P<path>{Str}) in base 64")
def step_impl(context, var_name, path):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    path = StepTools.evaluate_scenario_parameter(path)

    res = File.read_file_content_in_base64(path)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = content of file (?P<path>{Str}) in hexadecimal")
def step_impl(context, var_name, path):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    path = StepTools.evaluate_scenario_parameter(path)

    res = File.read_file_content_in_hexadecimal(path)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = lines of file (?P<path>{Str})")
def step_impl(context, var_name, path):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    path = StepTools.evaluate_scenario_parameter(path)

    with File(path, mode="r") as fin:
        lines = fin.readlines()
    res = list(map(lambda x: x.strip('\n'), lines))
    
    __get_variable_manager().register_variable(var_name, res)


@Given(r"(?P<var_name>{Variable}) = path to file with name (?P<filename>{Str})")
def step_impl(context, var_name, filename):
    var_name = StepTools.evaluate_variable_name(var_name)
    filename = StepTools.evaluate_scenario_parameter(filename)
    
    dest_path = __get_report_manager().current_scenario_report.get_path("files")
    res = os.path.join(dest_path, filename)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = path to timestamped file with prefix (?P<path_prefix>{Str}) and extension (?P<path_ext>{Str})")
def step_impl(context, var_name, path_prefix, path_ext):
    var_name = StepTools.evaluate_variable_name(var_name)
    prefix = StepTools.evaluate_scenario_parameter(path_prefix)
    ext = StepTools.evaluate_scenario_parameter(path_ext)

    res = __get_path_manager().get_timestamped_path(prefix, ext)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = create file with prefix (?P<path_prefix>{Str}) and extension (?P<path_ext>{Str}) and content (?P<content>{Str})")
def step_impl(context, var_name, path_prefix, path_ext, content):
    var_name = StepTools.evaluate_variable_name(var_name)
    execute_steps(u"""
        Given __FILENAME__@ = path to timestamped file with prefix {path_prefix} and extension {path_ext}
        Given {var_name} = create file with name __FILENAME__@ and content {content}
        """.format(path_prefix=path_prefix, path_ext=path_ext, var_name=var_name, content=content) )

@Given(r"(?P<var_name>{Variable}) = create file with name (?P<path_name>{Str}) and content (?P<content>{Str})")
def step_impl(context, var_name, path_name, content):
    var_name = StepTools.evaluate_variable_name(var_name)
    filename = StepTools.evaluate_scenario_parameter(path_name)
    content = StepTools.evaluate_scenario_parameter(content)
    
    dest_path = __get_report_manager().current_scenario_report.get_path("files")
    file_path = os.path.join(dest_path, filename)
    __get_path_manager().makedirs(file_path)
    
    File.write_file_with_content(file_path, content)

    __get_variable_manager().register_variable(var_name, file_path)


@Given(r"(?P<var_name>{Variable}) = create file with name (?P<path_name>{Str})")
def step_impl(context, var_name, path_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    content = BehaveStepTools.get_step_multiline_text(context, raise_exception_if_none=False)
    if content is None:
        content = ""
    __get_variable_manager().register_variable("__FILE_CONTENT__@", content)
    
    execute_steps(u"""
        Given {var_name} = create file with name {path_name} and content __FILE_CONTENT__@
        """.format(var_name=var_name, path_name=path_name) )


@Step(r"extract files from zip file (?P<zip_path>{Str}) in path (?P<dest_path>{Str})")
def step_impl(context, zip_path, dest_path):
    """
    extracts zip contents into the specified path.
    """
    zip_path = StepTools.evaluate_scenario_parameter(zip_path)
    dest_path = StepTools.evaluate_scenario_parameter(dest_path)
    
    __get_path_manager().makedirs(dest_path)
    
    # Extract files from the zip archive
    with ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(dest_path)
        logger.debug(f"Extracted files to {dest_path}")



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.verify_exception import VerifyException
import logging
from holado_scripting.text.verifier.functions.verify_function_extract_in import VerifyFunctionExtractIn
from holado_scripting.text.verifier.functions.verify_function_match_pattern import VerifyFunctionMatchPattern
from holado_scripting.text.base.text_inspecter import TextInspecter
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class TextVerifier(TextInspecter):

    def __init__(self):
        super().__init__()
        
        self.__text_interpreter = None
        self.__process_log_match_failure = True
        
    def initialize(self, var_manager, expression_evaluator, text_interpreter):
        super().initialize(var_manager, expression_evaluator)
        self.__text_interpreter = text_interpreter
        
        self.__register_default_functions()

    def verify(self, obtained, expected, do_check_case=True, raise_exception=True, do_log=True):
        res = True
        
        if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Verifying obtained '{obtained}' with expected '{expected}'...")
        
        # Check if expected is a verify function
        if expected is not None:
            result = self.__evaluate_function_verify(expected, obtained, raise_exception, do_log=do_log)
            if result is not None:
                return result
            
        # Interpret expected
        interpreted_expected = self.__text_interpreter.interpret(expected, log_level=logging.DEBUG if do_log else logging.TRACE)
        
        # Verify
        if interpreted_expected != obtained:
            # Build message
            msg = "Match failure: '{}' is different than '{}'".format(obtained, interpreted_expected)
            if interpreted_expected != expected:
                msg += " (from '{}')".format(expected)
                
            # log
            self._log_match_failure(msg, do_log=do_log)
            
            # Manage raise
            if raise_exception:
                raise VerifyException(msg)
            else:
                res = False
            
        if do_log and Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Verify obtained '{obtained}' with expected '{expected}' => [{res}]")
        return res

    def __evaluate_function_verify(self, expected, obtained, raise_exception, do_log):
        func_name, args_str = self._get_function_and_args_strings(expected)
        verify_function = self._get_function(func_name) if func_name is not None else None
        if verify_function is None:
            return None

        if func_name == "ExtractIn":
            args = [args_str]
        else:
            args = self._get_args(args_str)
            
        if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Evaluating verify function '{func_name}' with args {args} on obtained '{obtained}'")
        res = verify_function.apply(obtained, args, raise_exception=raise_exception)
        if do_log and Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Evaluate verify function [{expected}] => [{func_name}({', '.join(args)})] => [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res
                
    def __register_default_functions(self):
        self.register_function("MatchPattern", VerifyFunctionMatchPattern(self._variable_manager))
        self.register_function("ExtractIn", VerifyFunctionExtractIn(self._variable_manager))

    def _log_match_failure(self, msg, do_log=True):
        if self.__process_log_match_failure:
            if do_log and Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"--- VERIFIER --- {msg}")
        
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import time
import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_rabbitmq.tools.rabbitmq.rabbitmq_client import RMQClient,\
    RMQPublisher
import threading
import queue
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing


logger = logging.getLogger(__name__)


if RMQClient.is_available():
    import pika
    from pika import exceptions
    from pika.adapters.blocking_connection import _CallbackResult
    
    class RMQBlockingChannel(pika.adapters.blocking_connection.BlockingChannel):
        """
        Add features to pika.adapters.blocking_connection.BlockingChannel:
          - Rapid close: possibility to rapidly close connection without data consistency verification.
                         It is useful when using consumers for monitoring, like in testing context.
          - Add time_limit parameter to start_consuming method.
        """
        def __init__(self, channel_impl, connection, rapid_close=False):
            super().__init__(channel_impl, connection)
            self.__rapid_close = rapid_close
            self.__is_closing = False
            self.__confirm_delivery = False
            self.__is_consuming = False
            self.__is_stopping_consuming = False
        
        def close(self, reply_code=0, reply_text="Normal shutdown"):
            self.__is_closing = True
            try:
                super().close(reply_code=reply_code, reply_text=reply_text)
            except ValueError as exc:
                if self.__rapid_close and "file descriptor cannot be a negative integer" in str(exc):
                    # This exception can appear during channel close with rapid close option
                    pass
                else:
                    raise exc
        
        def _flush_output(self, *waiters):
            if self.__rapid_close and self.__is_closing:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Channel is closing, bypass output flush for rapid close")
            else:
                super()._flush_output(*waiters)
            
        def start_consuming(self, time_limit=None, time_limit_on_stop=None):
            """
            Override start_consuming method to add time_limit parameter.
            """
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self}] Start consuming...")
            self.__is_consuming = True
            try:
                if not self.__rapid_close and time_limit is None and time_limit_on_stop is None:
                    # Use pika implementation
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"[{self}] Start consuming with Pika implementation")
                    return super().start_consuming()
                else:
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"[{self}] Start consuming with HolAdo implementation")
                    
                # Check if called from the scope of an event dispatch callback
                with self.connection._acquire_event_dispatch() as dispatch_allowed:
                    if not dispatch_allowed:
                        raise exceptions.ReentrancyError(
                            'start_consuming may not be called from the scope of '
                            'another BlockingConnection or BlockingChannel callback')
                
                self._impl._raise_if_not_open()
                
                # Process events as long as consumers exist on this channel
                while self._consumer_infos:
                    # This will raise ChannelClosed if channel is closed by broker
                    if self.__is_stopping_consuming:
                        t_limit = time_limit_on_stop
                        if self.__rapid_close and t_limit is None:
                            # To ensure a rapid close, time_limit shouldn't be None, set it to 0 while stopping consuming
                            t_limit = 0
                    else:
                        t_limit = time_limit
                    if t_limit is None and self.__rapid_close:
                        # To ensure a rapid close, time_limit shouldn't be None, set it to 0.1s as a compromise between rapidity and resource cost 
                        t_limit = 0.1
                    
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[{self}] Processing data events (time_limit={t_limit})")
                    self._process_data_events(time_limit=t_limit)
            finally:
                self.__is_consuming = False
                self.__is_stopping_consuming = False
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[{self}] Finished start consuming")
        
        def stop_consuming(self, consumer_tag=None):
            self.__is_stopping_consuming = True
            super().stop_consuming(consumer_tag=consumer_tag)
        
        @property
        def has_confirm_delivery(self):
            return self._delivery_confirmation
        
        def set_confirm_delivery(self):
            self.confirm_delivery()
            


    class RMQBlockingConnection(pika.adapters.blocking_connection.BlockingConnection):
        """
        Add features to pika.adapters.blocking_connection.BlockingConnection:
          - Rapid close: possibility to rapidly close connection without data consistency verification.
                         It is useful when using consumers for monitoring, like in testing context.
        """
        def __init__(self, parameters=None, _impl_class=None, rapid_close=False):
            super().__init__(parameters, _impl_class)
            self.__rapid_close = rapid_close
            self.__is_closing = False

        def channel(self, channel_number=None):
            """
            Override channel method to use RMQBlockingChannel instead of BlockingChannel.
            The implementation is a copy of BlockingConnection.channel.
            """
            with _CallbackResult(self._OnChannelOpenedArgs) as opened_args:
                impl_channel = self._impl.channel(
                    channel_number=channel_number,
                    on_open_callback=opened_args.set_value_once)
    
                # Create our proxy channel
                channel = RMQBlockingChannel(impl_channel, self, rapid_close=self.__rapid_close)
    
                # Link implementation channel with our proxy channel
                impl_channel._set_cookie(channel)
    
                # Drive I/O until Channel.Open-ok
                channel._flush_output(opened_args.is_ready)
            
            return channel

        def close(self, reply_code=200, reply_text='Normal shutdown'):
            self.__is_closing = True
            super().close(reply_code=reply_code, reply_text=reply_text)
            
        def _flush_output(self, *waiters):
            if self.__rapid_close and self.__is_closing:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Connection is closing, bypass output flush for rapid close")
            else:
                super()._flush_output(*waiters)
                

class RMQBlockingClient(RMQClient):
    USE_RMQBlockingPublisher = True
    
    def __init__(self):
        super().__init__('RMQBlockingClient')
    
    def close(self):
        super().close()
        
        # Close all consumers connections
        for obj_kwargs in self._consumer_and_kwargs:
            connection = obj_kwargs.object.connection
            if connection and connection.is_open:
                self._close_connection(connection)
            
        for obj_kwargs in self._buffer_consumer_and_kwargs:
            connection = obj_kwargs.object.connection
            if connection and connection.is_open:
                self._close_connection(connection)
    
    def _new_connection(self):
        res = None
        connection_parameters = self._new_connection_parameters()
        
        last_exception = None
        for _ in range(300):
            try:
                res = RMQBlockingConnection(connection_parameters, rapid_close=self.rapid_close)
                break
            except pika.exceptions.IncompatibleProtocolError as exc:
                is_same_exception = str(last_exception) == str(exc) if last_exception else False
                if not last_exception or not is_same_exception:
                    logger.exception("Retry after possible temporary incompatible protocol errors while trying to connect with a blocking connection")
                    last_exception = exc
                elif is_same_exception:
                    logger.warning("Retry after same exception")
                time.sleep(0.1)
            except Exception as exc:
                is_same_exception = str(last_exception) == str(exc) if last_exception else False
                if not last_exception or not is_same_exception:
                    logger.exception("Retry after unexpected exception while trying to connect with a blocking connection")
                    last_exception = exc
                elif is_same_exception:
                    logger.warning("Retry after same exception")
                time.sleep(0.1)
                
        if res is None:
            raise TechnicalException(f"Failed to connect with a blocking connection and parameters [{connection_parameters}]. Last error was (more details in report): {last_exception} (type: {Typing.get_object_class_fullname(last_exception)})")
        return res
    
    def _new_publisher(self, connection, channel, exchange, routing_key, nb_runners):
        if self.USE_RMQBlockingPublisher:
            res = RMQBlockingPublisher(self, connection, channel, exchange, routing_key, nb_runners)
        else:
            res = RMQSimpleBlockingPublisher(self, connection, channel, exchange, routing_key)
        return res
        
    def new_consumer(self, queue, message_callback, queue_args=None, exchange="", exchange_args=None, bind_args=None):
        if not self.has_consumer and not self.has_buffer_consumer:
            # Use default connection for first consumer or buffer consumer
            return super().new_consumer(queue, message_callback, queue_args=queue_args, exchange=exchange, exchange_args=exchange_args, bind_args=bind_args)
        
        # Note: A new blocking connection is needed for each consumer due to multitasking possibilities of holado_rabbitmq.
        #       When a client hosts many consumers and start_consuming is called on it (RMQClient.start_consuming), 
        #       the method start_consuming is called in a separated thread for each consumer. But current implementation
        #       of BlockingChannel supposes that all pika methods are called in the same thread.
        connection = self._new_connection()
        return self._new_consumer_with_connection(connection, queue, message_callback, queue_args, exchange, exchange_args, bind_args)
    
    def new_buffer_consumer(self, queue, queue_args=None, exchange="", exchange_args=None, bind_args=None):
        if not self.has_consumer and not self.has_buffer_consumer:
            # Use default connection for first consumer or buffer consumer
            return super().new_buffer_consumer(queue, queue_args=queue_args, exchange=exchange, exchange_args=exchange_args, bind_args=bind_args)
        
        # Note: A new blocking connection is needed for each consumer due to multitasking possibilities of holado_rabbitmq.
        #       When a client hosts many consumers and start_consuming is called on it (RMQClient.start_consuming), 
        #       the method start_consuming is called in a separated thread for each consumer. But current implementation
        #       of BlockingChannel supposes that all pika methods are called in the same thread.
        connection = self._new_connection()
        return self._new_buffer_consumer_with_connection(connection, queue, queue_args=queue_args, exchange=exchange, exchange_args=exchange_args, bind_args=bind_args)
    
class RMQBlockingPublisher(RMQPublisher):
    """
    Implementation of a publisher for long and intensive publishes.
    """
    
    def __init__(self, client, connection, channel, exchange, routing_key, nb_runners=None):
        super().__init__(client, connection, channel, exchange, routing_key)

        self.__publish_lock = threading.Lock()
        self.__publish_nb_runners = nb_runners if nb_runners is not None else 10
        self.__publish_running_nb = 0
        self.__publish_queue = queue.Queue(maxsize=self.__publish_nb_runners * 2)
        self.__publish_counter = 0
        self.__published_counter = 0

    def flush(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] flushing...")
        self.__publish_queue.join()
        while self.__published_counter < self.__publish_counter:
            time.sleep(0.01)
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] flushed")

    def __publish_start(self):
        with self.__publish_lock:
            if self.__publish_running_nb < self.__publish_nb_runners:
                self.__publish_running_nb += 1
            else:
                return
        
        self.__publish_next()

    def __publish_next(self):
        try:
            msg_item = self.__publish_queue.get(block=False)
        except queue.Empty:
            with self.__publish_lock:
                self.__publish_running_nb -= 1
            return
        
        # Note: following code doesn't seem to be more efficient, thus it is commented
        # if self.client.is_in_connection_thread:
        #     self.client.connection.call_later(0, lambda: self.__publish_message(msg_item[0], **msg_item[1]))
        # else:
        #     self.client.connection.add_callback_threadsafe(
        #         lambda: self.__publish_message(msg_item[0], **msg_item[1])
        #         )
        self.client.connection.add_callback_threadsafe(
            lambda: self.__publish_message(msg_item[0], **msg_item[1]) )
        
        self.__publish_queue.task_done()

    def __publish_message(self, body, **kwargs):
        # Publish message
        RMQPublisher.publish(self, body, **kwargs)
        self.__published_counter += 1

        # Manage to publish next message
        self.__publish_next()

    def publish(self, body, **kwargs):
        # Start data events processing if needed
        self.client.start_process_data_events(raise_if_already_started=False)
        
        # Add message to publish in queue
        self.__publish_queue.put((body, kwargs))
        self.__publish_counter += 1

        # Start publishing if it is not already running
        self.__publish_start()

class RMQSimpleBlockingPublisher(RMQPublisher):
    """
    Simple implementation of a publisher
    """
    def __init__(self, client, connection, channel, exchange, routing_key):
        super().__init__(client, connection, channel, exchange, routing_key)
        


